/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "datatypes.h"

#include "namedparametertreeitem.h"

#include <cmath>
#include <QtCore/QMutex>
#include <QtCore/QMutexLocker>
#include <QtCore/QAbstractEventDispatcher>

#ifndef USESTATS
#define USESTATS 0 // 1=enabled 0=disabled
#endif

namespace NSCore {

	class MemoryPool
	{
		private:
			MemoryPool() {}
			MemoryPool( MemoryPool& ) {}
			MemoryPool( MemoryPool* ) {}
			~MemoryPool() {}

			struct MemEntry {
				double* mem;
				int size;
				MemEntry( double* m, int s ) : mem( m ), size( s ) {
				}
				MemEntry() : mem( 0 ), size( 0 ) {
				}
			};
			QMutex lock;
			QVector<MemEntry> activeChunks;
			QMultiMap<int,double*> chunkPool;

#if USESTATS
			struct Stats {
				int highest_in_use;
				int highest_in_pool;
				Stats() {
					highest_in_pool = 0;
					highest_in_use = 0;
				}
			} stats;
#endif
		public:
			static MemoryPool* the() {
				static MemoryPool* me = new MemoryPool();
				return me;
			}

			int chunkPoolSize() const {
				return chunkPool.size();
			}

			double* getMem( int size, bool allowhardwait = true ) {
				//qDebug() << "MemoryPool::getMem(" << size << ")";
				while ( allowhardwait && activeChunks.size() > 10000 ) {
					QAbstractEventDispatcher::instance()->processEvents( QEventLoop::AllEvents );
				}
				QMutexLocker l( &lock );
				double* tmp = 0;
				if ( chunkPool.count( size ) > 0 ) {
					tmp = chunkPool.take( size );
					//qDebug() << "MemoryPool::getMem() getting a chunk out of the pool";
				} else
					tmp = new double[ size ];
				activeChunks.push_back( MemEntry( tmp, size ) );
				//qDebug() << " Gave out a chunk. activeChunks.size() now returns" << activeChunks.size();
#if USESTATS
				stats.highest_in_use = qMax( stats.highest_in_use, activeChunks.size() );
#endif
				return tmp;
			}
			void freeMem( double* mem ) {
				QMutexLocker l( &lock );
				//qDebug() << "MemoryPool::freeMem(" << mem << ")";
				int my = -1;
				for ( int i=0; i<activeChunks.size(); ++i ) {
					if ( activeChunks[ i ].mem == mem )
						my = i;
				}
				//qDebug() << " This is chunk" << my << "out of" << activeChunks.size();
				if ( my > -1 ) {

					if ( chunkPoolSize() < 5000 ) {
						int size = activeChunks[ my ].size;
						chunkPool.insert( size, activeChunks[ my ].mem );
					} else
						delete[] activeChunks[ my ].mem;

					activeChunks.remove( my );
					//qDebug() << "Removed this chunk from the active chunks. (activeChunks.size() is now" << activeChunks.size() << ") chunkPool now contains" << chunkPool[ size ].size() << "for this size";
#if USESTATS
					stats.highest_in_pool = qMax( stats.highest_in_pool, chunkPool[ size ].size() );
#endif

				} else
					delete[] mem;
			}

#if USESTATS
			void debugPrint() const {
				qDebug() << "MemoryPool::debugPrint() highest in use:" << stats.highest_in_use;
				qDebug() << "MemoryPool::debugPrint() highest in pool:" << stats.highest_in_pool;
			}
#endif
	};

	ChannelDataSet::ChannelDataSet( int channel, int samples, AbstractTreeItem* parent )
		: AbstractTreeItem( QString( "channel %1" ).arg( channel ), parent )
		, _samples( samples )
		, _results( new NamedQVariantTreeItem( "results", QVariant::Invalid, this ) )
		, _parameters( new NamedQVariantTreeItem( "parameters", QVariant::Invalid, this ) )
	{
		//qDebug() << "ChannelDataSet::ChannelDataSet(" << channel << "," << samples << ") - new constructor";
		_parameters->addQVariant( "datatype", "acq-voltage" ); // Fullscale in voltage
		_parameters->addQVariant( "fullscale", 0.0 ); // Fullscale in voltage
		_parameters->addQVariant( "offset", 0.0 ); // Offset in voltage

		newGeneration();
		//qDebug() << "ChannelDataSet::ChannelDataSet() finished.";
	}
	ChannelDataSet::~ChannelDataSet() {
		//qDebug() << "ChannelDataSet::~ChannelDataSet()";
		while ( _data.size() > 0 )
			MemoryPool::the()->freeMem( _data.takeLast() );
		//qDebug() << " _data.size() is now" << _data.size();
#if USESTATS
		MemoryPool::the()->debugPrint();
#endif
	}

	void ChannelDataSet::addResult( QString path, QVariant value ) {
		//qDebug() << "ChannelDataSet::addResult(" << path << "," << value << ")";
		_results->addQVariant( QString( "generation %1/%2" ).arg( currentGeneration() ).arg( path ), value );
	}

	void ChannelDataSet::newGeneration() {
		//qDebug() << "ChannelDataSet::newGeneration()";
		_generations.push_back( _data.size() );
		_data.push_back( MemoryPool::the()->getMem( _samples, _data.size() == 0 ) );
		_min.push_back( INFINITY );
		_max.push_back( -INFINITY );
		addResult( "max", -INFINITY );
		addResult( "max at", -1 );
		addResult( "min", INFINITY );
		addResult( "min at", -1 );
		if ( _data.size() > 1 ) {
			//qDebug() << " Copying from previous generation";
			memcpy( _data[ _data.size() - 1 ], _data[ _data.size() - 2 ], _samples * sizeof( double ) );
			_min[ _min.size()-1 ] = _min[ _min.size()-2 ];
			_max[ _max.size()-1 ] = _max[ _max.size()-2 ];
			_results->getQVariant( QString( "generation %1/max" ).arg( _generations.back() ) )->setValue( result( "max", -1 )->toDouble() );
			_results->getQVariant( QString( "generation %1/max at" ).arg( _generations.back() ) )->setValue( result( "max at", -1 )->toInt() );
			_results->getQVariant( QString( "generation %1/min" ).arg( _generations.back() ) )->setValue( result( "min", -1 )->toDouble() );
			_results->getQVariant( QString( "generation %1/min at" ).arg( _generations.back() ) )->setValue( result( "min at", -1 )->toInt() );
		} else {
			//qDebug() << " creating blank first generation";
			for ( int i=0; i<_samples; i++ )
				_data.back()[ i ] = 0;
		}
		//qDebug() << " new number of generations is" << _generations.size() << _data.size() << _min.size() << _max.size();
		//qDebug() << "ChannelDataSet::newGeneration() finished";
	}
	void ChannelDataSet::popGeneration() {
		_generations.pop_back();
	}
	int ChannelDataSet::lastGeneration() const {
		if ( _generations.size() == 1 )
			return _generations.front();
		else {
			QList<int>::const_iterator it = _generations.end();
			--it;
			--it;
			return *it;
		}
	}
	int ChannelDataSet::currentGeneration() const {
		//qDebug() << "ChannelDataSet::currentGeneration() _generations.size() =" << _generations.size();
		return _generations.back();
	}
	void ChannelDataSet::setSample( int s, double value, int index ) {
		//qDebug() << "ChannelDataSet::setSample(" << s << "," << value << "," << index << ")";
		if ( index == -1000 )
			index = currentGeneration();
		//qDebug() << " modified index=" << index;
		if ( index == -1 )
			index = lastGeneration();
		//qDebug() << " modified index=" << index;
		_data[ index ][ s ] = value;
		if ( value > _max[ index ] ) {
			_results->getQVariant( QString( "generation %1/max" ).arg( index ) )->setValue( value );
			_results->getQVariant( QString( "generation %1/max at" ).arg( index ) )->setValue( s );
			_max[ index ] = value;
		}
		if ( value < _min[ index ] ) {
			_results->getQVariant( QString( "generation %1/min" ).arg( index ) )->setValue( value );
			_results->getQVariant( QString( "generation %1/min at" ).arg( index ) )->setValue( s );
			_min[ index ] = value;
		}
	}

	QVariant* ChannelDataSet::parameter( QString path ) const {
		QVariant* tmp =_parameters->getNamedChild<NamedQVariantTreeItem*>( QString( "parameters/%1" ).arg( path ) );
		if ( !tmp )
			qWarning() << "ChannelDataSet::parameter(" << path << ") will return 0";
		return tmp;
	}

	QVariant* ChannelDataSet::result( QString path, int index ) const {
		if ( index == -1000 )
			index = currentGeneration();
		if ( index == -1 )
			index = lastGeneration();
		NamedQVariantTreeItem* tmp = _results->getNamedChild<NamedQVariantTreeItem*>( QString( "results/generation %1/%2" ).arg( index ).arg( path ) );
		//if ( !tmp )
		//	qWarning() << "ChannelDataSet::result(" << path << "," << index << ") will return 0!";
		return tmp;
	}



	class AcqisitionDataSetPrivate : public AbstractTreeItem {
		public:
			int samples;

			AcqisitionDataSetPrivate() : AbstractTreeItem( "private", 0 ) {
				//qDebug() << "AcqisitionDataSetPrivate::AcqisitionDataSetPrivate()";
				samples = 0;
			}
			~AcqisitionDataSetPrivate() {
				//qDebug() << "AcqisitionDataSetPrivate::~AcqisitionDataSetPrivate()";
			}

			QAtomic ref;
		private:
			AcqisitionDataSetPrivate( const AcqisitionDataSetPrivate& other )
				: AbstractTreeItem( "private", 0 )
			{
				qDebug() << "AcqisitionDataSetPrivate::AcqisitionDataSetPrivate( other )";
				samples = other.samples;
				qDebug() << "AcqisitionDataSetPrivate::AcqisitionDataSetPrivate( other ) finished.";
			}
			QString debug() const { return "Private"; }
	};

	AcqisitionDataSet::AcqisitionDataSet( int channels, int samples )
	{
		d = new AcqisitionDataSetPrivate();
		d->ref.ref();
		d->samples = samples;
		NamedQVariantTreeItem* parameters = new NSCore::NamedQVariantTreeItem( "parameters", QVariant::Invalid, d );


		//qDebug() << "AcqisitionDataSet::AcqisitionDataSet(" << channels << "," << samples << ")";
		parameters->addQVariant( "samplinginterval", 0.0 ); // Sampling interval in seconds
		parameters->addQVariant( "triggerdelay", 0.0 ); // Trigger delay in seconds
		parameters->addQVariant( "timestamp", 0.0 ); // timestamp in picoseconds after start of the acquisation
		parameters->addQVariant( "horizontalposition", 0.0 ); // The time between -samplingintervall and 0 where the first sample occured

		for ( int i=0; i<channels; ++i ) {
			new ChannelDataSet( i, d->samples, d );
		}
		//qDebug() << " channels=" << channels << ": created" << _channels.size() << "channels";
		//qDebug() << " private d =" << d;
	}
	AcqisitionDataSet::AcqisitionDataSet( const AcqisitionDataSet& data )
		: d( data.d )
	{
		d->ref.ref();
		//qDebug() << endl << "AcqisitionDataSet::copy constructor ! this=" << this;
		//qDebug() << " private d =" << d;
		//qDebug() << "AcqisitionDataSet::copy constructor finished" << endl;
	}
	AcqisitionDataSet::~AcqisitionDataSet() {
		//qDebug() << "AcqisitionDataSet::~AcqisitionDataSet()";
		if ( ! d->ref.deref() ) {
			//qDebug() << " will delete d(" << d << ")";
			delete d;
		}
	}

	ChannelDataSet* AcqisitionDataSet::channel( int c ) const {
		//qDebug() << "AcqisitionDataSet::channel( int" << c << ")";
		ChannelDataSet* ret = d->getNamedChild<ChannelDataSet*>( QString( "channel %1" ).arg( c ) );
		if ( ret == 0 )
			qWarning() << "AcqisitionDataSet::channel( int" << c << ") will return 0!";
		return ret;
	}
	int AcqisitionDataSet::channels() const {
		//qDebug() << "AcqisitionDataSet::channels() returning" << _channels.size();
		if ( d && d->childs() > 0 )
			return d->childs()-1;
		return 0;
	}

	QVariant* AcqisitionDataSet::parameter( QString path ) const {
		return d->getNamedChild<QVariant*>( QString( "parameters/%1" ).arg( path ) );
	}

};

