/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DATATYPES_H
#define DATATYPES_H

#include <QtCore/QDebug>
#include <QtCore/QVariant>
#include <QtCore/QList>
#include <QtCore/QStringList>
#include <QtCore/QSharedDataPointer>

#include <vector>

#include <abstracttreeitem.h>
#include <namedparametertreeitem.h>
#include <namedqvarianttreeitem.h>

namespace NSCore {

	class AcqisitionDataSet;

	/**
	 * @brief Data for one channel
	 *
	 * This class holds the data from one channel.
	 *
	 * It has a number of samples and any number and kind of results can be
	 * attached to it.
	 *
	 * Current set of parameters is:
	 * <ul>
	 * <li>datatype - which gets set to "acq-voltage" and means that the data is
	 * double and 1 is 1V
	 * <li>fullscale - the fullscale of this channel
	 * <li>offset - the offset the fullscale was shifted
	 * </ul>
	 *
	 * It also features the idea of data-generations which act like overlays
	 * or different versions of the data.
	 * @see newGeneration(), popGeneration(), currentGeneration(),
	 * lastGeneration()
	 *
	 * @todo sort the generations into a tree?
	 */
	class ChannelDataSet : public NSCore::AbstractTreeItem {
		public:
			/**
			 * @brief constructor
			 *
			 * Creates a new channel with the given number of samples.
			 *
			 * @param channel the number of the channel, the own name will then
			 * be "channel %i"
			 * @param samples the number of samples this set holds
			 * @param parent the parent AbstractTreeItem, which should be an
			 * AcqisitionDataSet
			 */
			ChannelDataSet( int channel, int samples, AbstractTreeItem* parent );

			/// Destructor
			~ChannelDataSet();

		public:
			/**
			 * @brief Returns the number of samples in this ChannelDataSet
			 */
			int samples() const { return _samples; }
			/**
			 * @brief Returns the desired sample
			 *
			 * @param s The number of the sample to return
			 * @param index The index of the generation to return the sample
			 * from. The default ( -1000 ) means the currentGeneration(), -1
			 * means lastGeneration().
			 */
			double sample( int s, int index=-1000 ) const {
				if ( index == -1000 )
					index = currentGeneration();
				if ( index == -1 )
					index = lastGeneration();
				if ( s < 0 || s >= _samples )
					return 0;
				return _data[ index ][ s ];
			}
			/**
			 * @brief Returns a pointer to the double-array that is the sample
			 * data
			 *
			 * This can be used for functions that work directly on arrays. It
			 * can also be used to do memcopies. But be sure not to delete the
			 * data!
			 *
			 * @param index The index of the data generation to return. The
			 * default ( -1000 ) means currentGeneration(), -1 means
			 * lastGeneration().
			 */
			double* sampledata( int index=-1000 ) {
				if ( index == -1000 )
					index = currentGeneration();
				if ( index == -1 )
					index = lastGeneration();
				return _data[ index ];
			}
			/**
			 * @brief Set the desired sample to a new value
			 *
			 * @param s the number of the sample to return
			 * @param value the value the sample should get
			 * @param index the index of the generation to return the sample
			 * from. The default ( -1000 ) means the currentGeneration(), -1
			 * means lastGeneration().
			 */
			void setSample( int s, double value, int index=-1000 );

			/**
			 * @brief return a specific parameter
			 */
			QVariant* parameter( QString path ) const;

			/**
			 * @brief Add a result with its path
			 */
			void addResult( QString path, QVariant value );

			/**
			 * @brief gives access to the tree of results
			 *
			 * This returns a pointer to the whole result-tree. Maybe it is
			 * wise in the future to also create a read-only interface of
			 * NamedQVariantTreeItem to give read-only access to the data...
			 */
			NamedQVariantTreeItem* results() const { return _results; }

			/**
			 * @brief Return a specific result
			 */
			QVariant* result( QString path, int index=-1000) const;

			/**
			 * @brief returns the number of data-generations available
			 *
			 * This function is needed for OsciWidget so it doesn't try to show
			 * non-existant generations.
			 */
			int generations() const { return _data.size(); }
			/**
			 * @brief Create a new Generation
			 *
			 * creates a new generation of data and sets currentGeneration to it
			 */
			void newGeneration();
			/**
			 * @brief Go back one generation
			 */
			void popGeneration();
			/**
			 * @brief returns the current generation
			 */
			int currentGeneration() const;
			/**
			 * @brief Return the last generation
			 */
			int lastGeneration() const;

		private:
			/**
			 * @brief The normal disabled copy-constructor
			 *
			 * This has to be private as there is no possibility to copy it
			 * without a pointer to the new parent.
			 */
			ChannelDataSet( const ChannelDataSet& )
				: AbstractTreeItem( "bla", 0 )
				, _results( 0 )
				, _parameters( 0 )
			{
			}
			ChannelDataSet( ChannelDataSet* )
				: AbstractTreeItem( "bla", 0 )
				, _results( 0 )
				, _parameters( 0 )
			{
			}

			QString debug() const {
				return QString( "ChannelDataSet (%1) %2" )
					.arg( name() ).arg( path() );
			}

			QList <double*> _data;
			int _samples;
			NamedQVariantTreeItem* _results;
			NamedQVariantTreeItem* _parameters;

			QList<int> _generations;
			/**
			 * @{
			 *
			 * @brief Caching
			 *
			 * These values are normally accessible via the results() but cached
			 * here for faster usage in setSample().
			 */
			QList<double> _min;
			QList<double> _max;
			// @}
	};

	class AcqisitionDataSetPrivate;

	/**
	 * @brief New generation acquisation data...
	 *
	 * You can savely create copies of objects of this type, all the data is
	 * saved in pointers with refcounting. Thus all copies are shallow.
	 *
	 * Current parameters are:
	 * <ul>
	 * <li>samplinginterval - The time between two samples
	 * <li>triggerdelay - the time the trigger is shifted from the beginning
	 * <li>timestamp - some timestamp the acqiris-digitizers present
	 * <li>horizontalposition - the horizontal shifting of the first sample with
	 * respect to the trigger
	 * </ul>
	 *
	 * @see parameter()
	 *
	 * @todo Results per dataset. This probably means that generations are
	 * handled at this level and not in ChannelDataSet
	 */
	class AcqisitionDataSet
	{
		public:
			/**
			 * @brief creates a new AcqisitionDataSet
			 */
			AcqisitionDataSet( int channels, int samples );
			/**
			 * @brief copy-constructor
			 */
			AcqisitionDataSet( const AcqisitionDataSet& );
			/// destructor
			~AcqisitionDataSet();

			/**
			 * @brief returns a specific channel
			 *
			 * If there is no channel with the desired number this returns 0.
			 */
			ChannelDataSet* channel( int ) const;
			/**
			 * @brief returns the number of channels
			 */
			int channels() const;

			/**
			 * @brief return a specific parameter as QVariant
			 *
			 * This gives access only to the direct parameters of the
			 * AcqisitionDataSet, not of the ChannelDataSets.
			 */
			QVariant* parameter( QString ) const;

		private:
			AcqisitionDataSetPrivate* d;

	};

};

#endif // DATATYPES_H

