/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "gaussproducer.h"
#include "gaussproducer.moc"

#include <cmath>
#include <cstdlib>
#include <QtCore/QDateTime>
#include <QtCore/QTimer>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>

#include <actiontreeitem.h>

using namespace NSCore;

// Get the randoms in gauss-distribution from the polar-method (faster than box-muller)
// see http://de.wikipedia.org/wiki/Normalverteilung#Simulation_normalverteilter_Zufallsvariablen
double polarmethodrandom() {
	double u1, u2;
	double v = 2;
	while ( v >= 1 ) {
		u1 = rand()*1.0/RAND_MAX;
		u2 = rand()*1.0/RAND_MAX;
		v = pow( 2*u1-1, 2 ) + pow( 2*u2-1, 2 );
	}
	return ( 2*u1-1 )*sqrt( -2*log( v )/v );
}

GaussProducer::GaussProducer( QObject* p )
	: DataSetProducer( p )
	, _togenerate( 0 )
	, _p( new NamedParameterTreeItem( "Parameters", QVariant::Invalid, this ) )
	, _dirty( true )
{
	_cache.heights = new double[ 2 ];
	_cache.heightsigmas = new double[ 2 ];
	setObjectName( "Gauss-Producer" );
	srand( QDateTime::currentDateTime().toTime_t() );
	_ownthread = ( thread() != QCoreApplication::instance()->thread() );

	connect( _p, SIGNAL( valueChanged() ), this, SLOT( cacheDirty() ) );
	_p->addParameter( "Samples", 100 );
	_p->getParameter( "Samples" )->setRange( 0, 1024 );
	_p->addParameter( "Resolution", QVariant( 0 ) );
	_p->getParameter( "Resolution" )->setRange( 0, 64 );
	_p->getParameter( "Resolution" )->setToolTip( "Special value 0 means to use double-precision" );
	_p->addParameter( "Noise Level", 0.1 );
	_p->getParameter( "Noise Level" )->setRange( 0.0, 100.0 );
	_p->addParameter( "Sigma", 2.5 );
	_p->getParameter( "Sigma" )->setRange( 0.0, 100.0 );
	_p->addParameter( "Gauss-Sigma", 0.5 );
	_p->getParameter( "Gauss-Sigma" )->setRange( 0.0, 100.0 );
	_p->addParameter( "Channel 1/Height", 1.1 );
	_p->addParameter( "Channel 1/Sigma", 0.2 );
	_p->addParameter( "Channel 2/Height", 0.5 );
	_p->addParameter( "Channel 2/Sigma", 0.2 );

	new ActionTreeItem( "Generate one event!", this, ":/icons/kde/cr22-action-player_play.png", this, SLOT( generateOneDataSet() ) );
	ActionTreeItem* action = new ActionTreeItem( "Generate N events!", this, ":/icons/kde/cr22-action-player_fwd.png", this, SLOT( generateNDataSets() ) );
	_number = new NamedParameterTreeItem( "Events to generate", 1, action );
}
GaussProducer::~GaussProducer() {
}

void GaussProducer::generateNDataSets( int n ) {
	//qDebug() << "GaussProducer::generateNDataSets(" << n << ")";
	if ( _togenerate == 0 )
		QTimer::singleShot( 100, this, SLOT( generateDataSet() ) );
	_togenerate += n;
}

QString GaussProducer::toXML() const {
	//return QString( "<reader type=\"gaussproducer\" samples=\"%1\" resolution=\"%2\" samplinginterval=\"%3\" noiselevel=\"%4\" sigma=\"%5\" gausssigma=\"%6\" />" ).arg( samples() ).arg( resolution() ).arg( samplingInterval() ).arg( noiseLevel() ).arg( sigma() ).arg( gaussSigma() );
	return QString( "<reader type=\"gaussproducer\">%1</reader>" ).arg( _p->toXML() );
}

void GaussProducer::fromXML( QString xml ) {
	QDomDocument doc( "gausproducer-doc" );
	doc.setContent( xml );
	fromXML( doc.documentElement() );
}

void GaussProducer::fromXML( QDomNode node ) {
	//qDebug() << "GaussProducer::fromXML(" << &node << ")";
	if ( node.nodeName() != "reader" )
		return;

	QDomNamedNodeMap items = node.attributes();
	if ( items.contains( "type" ) ) {
		if ( items.namedItem( "type" ).toAttr().value() != "gaussproducer" )
			return;
	}

	//qDebug() << "Seems to be a reader-item. Will continue...";

	if ( items.contains( "samples" ) )
		_p->getParameter( "Samples" )->setValue( items.namedItem( "samples" ).toAttr().value().toInt() );
	if ( items.contains( "resolution" ) )
		_p->getParameter( "Resolution" )->setValue( items.namedItem( "resolution" ).toAttr().value().toInt() );
	if ( items.contains( "noiselevel" ) )
		_p->getParameter( "Noise Level" )->setValue( items.namedItem( "noiselevel" ).toAttr().value().toDouble() );
	if ( items.contains( "sigma" ) )
		_p->getParameter( "Sigma" )->setValue( items.namedItem( "sigma" ).toAttr().value().toDouble() );
	if ( items.contains( "gausssigma" ) )
		_p->getParameter( "Gauss-Sigma" )->setValue( items.namedItem( "gausssigma" ).toAttr().value().toDouble() );

	QDomNode options = node.firstChildElement( "parameter" );
	NSCore::NamedParameterTreeItem::setValueFromNode( _p, options );
}

void GaussProducer::generateDataSet() {

	if ( QThread::currentThread() != thread() ) {
		QTimer::singleShot( 10, this, SLOT( generateDataSet() ) );
		return;
	}

	if ( _dirty ) {
		_cache.samples = _p->getParameter( "Samples" )->toInt();
		_cache.bits = _p->getParameter( "Resolution" )->toInt();
		_cache.noise = _p->getParameter( "Noise Level" )->toDouble();
		_cache.sigma = _p->getParameter( "Sigma" )->toDouble();
		_cache.gausssigma = _p->getParameter( "Gauss-Sigma" )->toDouble();

		for ( int i=0; i<2; ++i ) {
			_cache.heights[ i ] = _p->getParameter( QString( "Channel %1/Height" ).arg( i+1 ) )->toDouble();
			_cache.heightsigmas[ i ] = _p->getParameter( QString( "Channel %1/Sigma" ).arg( i+1 ) )->toDouble();
		}

		_dirty = false;
	}

	//qDebug() << "GaussProducer::generateDataSet()";

	const int channels = 2;

	AcqisitionDataSet* data = new AcqisitionDataSet( channels, _cache.samples );

	const double offset = 0;

	const int fullrange = int( pow( 2, _cache.bits ) );

	double pos = 0;

	//qDebug() << "Have to generate data for" << data.channels() << "channels.";

	for ( int i=0; i<data->channels(); ++i ) {
		//qDebug() << "Generating data for channel" << i;

		// Move the first pulse a random amount and the second one relative to the first one
		if ( i==0 )
			pos = rand()*4.0/RAND_MAX - 2;
		else {
			double p = _cache.gausssigma * polarmethodrandom();
			pos += p;
			data->parameter( "horizontalposition" )->setValue( p );
		}

		// Cache to be faster...
		ChannelDataSet* channel = data->channel( i );

		double noise = 0;

		double impulseheight = polarmethodrandom() * _cache.heightsigmas[ i ] + _cache.heights[ i ];

		for ( int j=0; j<channel->samples(); ++j ) {
			noise = ( rand()*2.0/RAND_MAX - 1 ) * _cache.noise;
			double v =- exp( - pow( j-_cache.samples/2-pos, 2 ) / ( 2* pow( _cache.sigma, 2 ) ) ) *impulseheight + noise + offset;
			if ( _cache.bits != 0 )
				v = int( v * 0.5 * fullrange ) * 2.0 / fullrange;
			channel->setSample( j, v );
		}

		channel->parameter( "fullscale" )->setValue( 3.0 );
		channel->parameter( "offset" ) ->setValue( 1.4 - offset );
	}

	enqueue( data );

	--_togenerate;
	if ( _togenerate > 0 ) {
		if ( _ownthread )
			QTimer::singleShot( 0, this, SLOT( generateDataSet() ) );
		else
			QTimer::singleShot( 10, this, SLOT( generateDataSet() ) );
	}
}

