/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef GAUSSPRODUCER_H
#define GAUSSPRODUCER_H

#include <QtCore/QObject>
#include "datasetchain.h"
#include "namedparametertreeitem.h"

namespace NSCore {

	/**
	 * @brief Produces Gauss-shaped pulses with a delay that is gauss-distributed.
	 *
	 * This class can ( and will ) be used to determine the mathematical resolution of algorithms or to simulate and test other parts of the system.
	 *
	 * It produces a pair of pulses that are gauss-shaped and that have a time-distance that is gauss-distributed. Several options can be changed:
	 * - The number of samples in a dataset.
	 * - The bit-resolution ( special: 0 means double ).
	 * - The samplinginterval and the sigma of the gausspulse to generate.
	 * - The noiselevel.
	 * - The sigma of the delaying gauss-function.
	 */
	class GaussProducer : public DataSetProducer
	{
		Q_OBJECT
		public:
			/// Simple constructor
			GaussProducer( QObject* =0 );
			/// Simple destructor
			~GaussProducer();

			/**
			 * @brief Save all the settings to XML
			 */
			QString toXML() const;
			/**
			 * @brief Load settings from XML
			 *
			 * @{
			 */
			void fromXML( QString );
			void fromXML( QDomNode );
			// @}

		public slots:

			/**
			 * @brief Generate a number of datasets
			 */
			void generateNDataSets( int );

			/// Empty stub to make the eode-macros happy
			void setFile( QString ) {};

		private slots:
			/**
			 * @brief The real working function
			 */
			void generateDataSet();

			/**
			 * @brief mark the internal parameter-cache as dirty.
			 * @internal
			 */
			void cacheDirty() { _dirty = true; }

			void generateOneDataSet() { generateNDataSets( 1 ); }
			void generateNDataSets() { generateNDataSets( _number->toInt() ); }

		private:
			QString debug() const { return "GaussProducer"; }

			int _togenerate;
			bool _ownthread;

			NamedParameterTreeItem* _p;

			NamedParameterTreeItem* _number;

			bool _dirty;
			struct cache_t {
				int samples, bits;
				double noise;
				double sigma;
				double gausssigma;
				double *heights;
				double *heightsigmas;
			} _cache;
	};

};

#endif // GAUSSPRODUCER_H

