/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef MATHEXPRESSION_H
#define MATHEXPRESSION_H

#include <abstracttreeitem.h>
#include <datatypes.h>

namespace NSCore {

/**
 * @brief Baseclass for an expression
 *
 * Expressions are used to extract any kind of results from the
 * NSCore::AcqisitionDataSet respectively from the results saved in them.
 */
class MathExpression : public NSCore::QTreeItem
{
	public:
		/// simple tree-constructor
		MathExpression( NSCore::QTreeItem* parent );

		QString guiString() const;
		QStringList guiActions() const;

		void guiExecAction( QString name, NSCore::GuiServer_Interface* gui );

		/**
		 * @brief delete items in a thread-save way...
		 *
		 * ...by handing the action to the parent until it gets handled in a
		 * different way.
		 */
		virtual void deleteMe( MathExpression* exp, NSCore::GuiServer_Interface* gui );
		/**
		 * @brief add new items in a thread-save way...
		 *
		 * ...by handing the action to the parent until it gets handled in a
		 * different way.
		 */
		virtual void addChild( MathExpression* pexp, NSCore::GuiServer_Interface* gui );

		/**
		 * @brief Abstract function that does the work
		 *
		 * This has to be implemented to do the work.
		 */
		virtual double value( NSCore::AcqisitionDataSet& ) =0;

		/**
		 * @brief save everything needed to recreate the path from xml
		 */
		virtual QString toXML() const =0;
};


/**
 * @brief Singleton to create MathExpression trees
 *
 * This MathExpressionFromXMLCreator is to provide one ( extendible ) way of
 * creating MathExpressions or maybe even QTreeItems. Basicly all the knowledge
 * to create MathExpressions is centralized here to provide a single api for all
 * users.
 *
 * It should be extendible in a way that derived creators register themself at
 * the global creator and are then used to create their kind of items.
 */
class MathExpressionFromXMLCreator
{
	public:
		/**
		 * @brief Access to the global instance
		 */
		static MathExpressionFromXMLCreator* the();

		/**
		 * @brief Create MathExpressions from xml
		 *
		 * Either from a QString or from a QDomNode
		 *
		 * @{
		 */
		MathExpression* fromXML( QString, QTreeItem* parent );
		MathExpression* fromXML( QDomNode, QTreeItem* parent );
		// @}

		/**
		 * @brief Return a list of all the elements creatable
		 *
		 * This asks all the registered creators for their list of elements and returns the combination
		 *
		 * @see elementsICanCreate
		 */
		QStringList availableElements() const;
		/**
		 * @brief Return a helptext for a certain type of element
		 *
		 * This can be used for GUI to easen the decision.
		 */
		QString helpText( QString ) const;

	protected:
		/**
		 * @brief default-constructor
		 *
		 * Registers itself at the global list of creators
		 */
		MathExpressionFromXMLCreator();
		/**
		 * @brief destructor
		 *
		 * Unregisters itself from the global list of creators
		 */
		virtual ~MathExpressionFromXMLCreator() {
			_creators.removeAll( this );
		}

		/**
		 * @brief Construct a single element from xml
		 */
		virtual MathExpression* singleElementFromXML( QDomNode, QTreeItem* pexp );

		/**
		 * @brief Return the string list of elements to create
		 */
		virtual QStringList elementsICanCreate() const;

		/**
		 * @brief Return the help text for a certain element-type
		 */
		virtual QString myHelpText( QString ) const;

	private:
		/**
		 * @brief No copying allowed.
		 */
		MathExpressionFromXMLCreator( MathExpressionFromXMLCreator& ) {}

		/**
		 * @internal
		 *
		 * @brief Stupid constructor to be used by the()
		 *
		 * This one doesn't register itself at the global list of creators,
		 * doing so inside of the() would result in an endless loop.
		 */
		MathExpressionFromXMLCreator( bool ) {}

		/**
		 * @brief List of all creators
		 */
		QList<MathExpressionFromXMLCreator*> _creators;
};

/**
 * @brief Simple result
 *
 * This returns a simple result. The parameters are the name, generation and
 * channel of the result to fetch.
 */
class MathExpression_Result : public MathExpression
{
	public:
		MathExpression_Result( QString name, int channel, int generation, NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString name() const { return _name->toString(); }
		int channel() const { return _channel->toInt(); }
		int generation() const { return _generation->toInt(); }

		QString toXML() const;
	private:
		QString debug() const { return QString( "result: %1" ).arg( _name->toString() ); }
		NSCore::NamedParameterTreeItem *_name, *_channel, *_generation;
};

/**
 * @brief Parameter of the sample
 *
 */
class MathExpression_Parameter : public MathExpression
{
	public:
		MathExpression_Parameter( QString name, int channel, NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString name() const { return _name->toString(); }
		int channel() const { return _channel->toInt(); }

		QString toXML() const;
	private:
		QString debug() const { return QString( "result: %1" ).arg( _name->toString() ); }
		NSCore::NamedParameterTreeItem *_name, *_channel;
};

/**
 * @brief Simply a constant
 *
 */
class MathExpression_Constant : public MathExpression
{
	public:
		MathExpression_Constant( double value, NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString toXML() const;
	private:
		QString debug() const { return QString( "constant: %1" ).arg( _value->toString() ); }
		NSCore::NamedParameterTreeItem *_value;
};

/**
 * @brief Addition of results
 *
 * This adds the child results together. There are no parameters for this.
 */
class MathExpression_Add : public MathExpression
{
	public:
		MathExpression_Add( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "add"; }

		QStringList guiActions() const { return MathExpression::guiActions() << "Add Exp."; }

		QString toXML() const;
};

/**
 * @brief Multiplication of results
 *
 * This multiplies the child results together. There are no parameters for this.
 */
class MathExpression_Mul : public MathExpression
{
	public:
		MathExpression_Mul( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "mul"; }

		QStringList guiActions() const { return MathExpression::guiActions() << "Add Exp."; }

		QString toXML() const;
};

/**
 * @brief The minimum of a number of values
 */
class MathExpression_Min : public MathExpression
{
	public:
		MathExpression_Min( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "min"; }

		QStringList guiActions() const { return MathExpression::guiActions() << "Add Exp."; }

		QString toXML() const;
};
/**
 * @brief The maximum of a number of values
 */
class MathExpression_Max : public MathExpression
{
	public:
		MathExpression_Max( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "max"; }

		QStringList guiActions() const { return MathExpression::guiActions() << "Add Exp."; }

		QString toXML() const;
};


/**
 * @brief Negate the result
 *
 * This negates the childs result. If there are multiple childs, only the last
 * result will be used.
 */
class MathExpression_Neg : public MathExpression
{
	public:
		MathExpression_Neg( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "neg"; }

		QString guiString() const { return "Negate ( * -1 )"; }

		QStringList guiActions() const { return MathExpression::guiActions() << "Add Exp."; }

		QString toXML() const;
};

/**
 * @brief Reduce the precision of a result
 *
 * This reduces the precision of the result so far. The given parameter is the precision which can be any number to which the result is trimmed to. For example 5E-2 will give you results with a precision of 0.05.
 *
 * If there are multiple childs, only the last result will be used.
 */
class MathExpression_Precision : public MathExpression
{
	public:
		MathExpression_Precision( double precision, NSCore::QTreeItem* p );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return QString( "precision: %1" ).arg( _precision->toDouble() ); }

		QStringList guiActions() const { return MathExpression::guiActions() << "Add Exp."; }

		QString toXML() const;
	private:
		NSCore::NamedParameterTreeItem *_precision;
};

/**
 * @brief Rather complicated SmallerMinusBigger
 *
 * This one compares a result on the first two channels to select the bigger and
 * the smaller one and then substractes a second kind of result of both
 * channels. Which channels result gets subtracted from the other is selected by
 * the first decision.
 */
class MathExpression_SmB : public MathExpression
{
	public:
		MathExpression_SmB( QString selection, QString calculus, int generation, NSCore::QTreeItem* p, int generation_calc = -2000 );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return QString( "Smaller-Minus-Bigger: selection='%1' calculus='%2' generation_select='%3' generation_calc='%4'" ).arg( _selection->toString() ).arg( _calculus->toString() ).arg( _generation_select->toString() ).arg( _generation_calc->toString() ); }

		QString guiString() const { return "Smaller-Minus-Bigger"; }

		QString selection() const { return _selection->toString(); }
		QString calculus() const { return _calculus->toString(); }
		int generation() const { return _generation_select->toInt(); }

		QString toXML() const;

	private:
		NSCore::NamedParameterTreeItem *_selection, *_calculus, *_generation_select, *_generation_calc;
};

/**
 * @brief Greater-than-Zero
 */
class MathExpression_GtZ : public MathExpression
{
	public:
		MathExpression_GtZ( NSCore::QTreeItem* );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "Greater-than-Zero"; }
		QString guiString() const { return "Greater-than-Zero"; }

		QStringList guiActions() const { return MathExpression::guiActions() << "Add Exp."; }

		QString toXML() const;
};
/**
 * @brief Less-than-Zero
 */
class MathExpression_LtZ : public MathExpression
{
	public:
		MathExpression_LtZ( NSCore::QTreeItem* );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "Less-than-Zero"; }
		QString guiString() const { return "Less-than-Zero"; }

		QStringList guiActions() const { return MathExpression::guiActions() << "Add Exp."; }

		QString toXML() const;
};
/**
 * @brief Pass the value if its inside the window
 */
class MathExpression_Window : public MathExpression
{
	public:
		MathExpression_Window( double min, double max, NSCore::QTreeItem* );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "Window"; }
		QString guiString() const { return "Window"; }

		QStringList guiActions() const { return MathExpression::guiActions() << "Add Exp."; }

		QString toXML() const;

	private:
		NamedParameterTreeItem *_min, *_max;
};

}; // NSCore

#endif // MATHEXPRESSION_H

