/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef QTREEITEM_H
#define QTREEITEM_H

#include <QtCore/QObject>
#include <QtCore/QStringList>
#include <QtCore/QDebug>
#include <QtCore/QVariant>

#include <pointer_template.h>
#include <guiserver_interface.h>

#if ( QT_POINTER_SIZE == 8 )
#define pint qint64
#else
#define pint qint32
#endif

namespace NSCore {

	class QTreeItem;

	/**
	 * @brief A pointer to QTreeItem
	 *
	 * This is mostly an extension to the pointer-template but introduces two
	 * new functions to derive of.
	 */
	class QTreeItemPointer : public Pointer_template<QTreeItem>
	{
		public:
			/// Constructor
			QTreeItemPointer( QTreeItem* p );

			void setNull() {
				itemDeleted();
				Pointer_template<QTreeItem>::setNull();
			}
			/**
			 * @brief The item pointed to is deleted.
			 *
			 * Implement this functions if you want to be notified of the
			 * deletion of the item pointed to. Better not use the item anymore,
			 * with threads it can be that the item is already deleted.
			 */
			virtual void itemDeleted() {}
			/**
			 * @brief A new child was added.
			 *
			 * This signals that a new child was added to the item pointed to.
			 * Can be used to create the new pointer to the new child.
			 */
			virtual void newChildAdded( QTreeItem* ) {}
	};

	/**
	 * @brief Abstract class to arrange items in a tree.
	 *
	 * This class takes care of accessing the items and numbering them.
	 *
	 * <div style="text-decoration:line-through;">An QTreeItem can have a
	 * maximum number of childs set, so adding one to many might fail.</div>
	 * This is currently not implemented. If it is really needed it can come
	 * back but I don't see why.
	 *
	 * Managing childs is mainly done by QObject. But as we have the opportunity
	 * to add Childs which don't belong to this QObject, please add the foreign
	 * childs with addChild(...) and use childobjects() to retrieve the
	 * childrens instead of QObject::children().
	 */
	class QTreeItem : public QObject, public PointerProtection_Skel {
		Q_OBJECT
		public:
			/**
			 * @brief Create a new QTreeItem
			 *
			 * Be aware that if the parent isn't 0 and is a QTreeItem it adds
			 * itself to the parents childs. So no further action has to be done.
			 *
			 * @param p The parent QTreeItem or QObject this one got inserted to.
			 */
			QTreeItem( QObject* p=0 )
				: QObject( p )
				, _deferredDelete( false )
			{
				//qDebug() << "QTreeItem::QTreeItem("<< p <<")";
				if ( p!=0 ) {
					QTreeItem* parent = dynamic_cast<QTreeItem*>( p );
					if ( parent )
						parent->addChild( this );
				}
			}
			/**
			 * @brief Create a new named QTreeItem
			 *
			 * Be aware that if the parent isn't 0 and is a QTreeItem it adds
			 * itself to the parents childs. So no further action has to be done.
			 *
			 * @param n The name of this QTreeItem. Allows accessing this item 
			 * and childs by name / path.
			 * @param p The parent QTreeItem or QObject this one got inserted to.
			 */
			QTreeItem( QString n, QObject* p=0 )
				: QObject( p )
				, _deferredDelete( false )
			{
				setObjectName( n );
				//qDebug() << "QTreeItem::QTreeItem("<< p <<")";
				if ( p!=0 ) {
					QTreeItem* parent = dynamic_cast<QTreeItem*>( p );
					if ( parent )
						parent->addChild( this );
				}
			}
			/**
			 * @brief Destructor
			 */
			virtual ~QTreeItem() {
				//qDebug() << "QTreeItem::~QTreeItem()";
			};

			/// Return the number of childs.
			int childs() const { return childobjects().size(); }

			/**
			 * @brief Return pointers to all the child-treeitems.
			 *
			 * This extends QObject::children() in that it also returns childs
			 * that are not children in the QObject-way.
			 *
			 * @note For usage on QTreeItem's tree-structure this has to be
			 * used instead of the QObject::children()!
			 */
			const QObjectList childobjects() const {
				QObjectList list = children();
				foreach( QObject* obj, _morechilds )
					list.append( obj );

				foreach( QObject* obj, list ) {
					QTreeItem* item = dynamic_cast<QTreeItem*>( obj );
					//if ( !item )
					//	qWarning() << "QTreeItem::childobjects() have a child here that is not a QTreeItem*!";
					if ( item && item->_deferredDelete ) {
						//qDebug() << "QTreeItem::childobjects() removing one item that is in deferred delete state!";
						list.removeAll( obj );
					}
				}

				return list;
			}

			/**
			 * @brief Return pointers to all child-treeitems of a certain type
			 *
			 * This extends childobjects() in that it only returns the addresses
			 * of the direct children that are of the given type.
			 */
			template<class T>
			QList<T*> childobjects() const {
				QList<T*> ret;
				foreach( QObject* obj, childobjects() ) {
					T* tmp = dynamic_cast<T*>( obj );
					if ( tmp )
						ret.append( tmp );
				}
				return ret;
			}

			/**
			 * @brief The names of the direct childs.
			 *
			 * This returns a QStringList with the names of the direct children.
			 */
			const QStringList childlist() const {
				QStringList res;
				foreach( QObject* child, childobjects() ) {
					res << child->objectName();
				}
				return res;
			}

			/**
			 * @brief Gives access to the childs by their number.
			 *
			 * @param n the child you want to access
			 */
			QTreeItem* child( int n ) const {
				QList<QObject*> list = childobjects();
				if ( n<list.size() && dynamic_cast<QTreeItem*>( list[ n ] ) )
					return dynamic_cast<QTreeItem*>( list[ n ] );
				else
					return 0;
			}

			/**
			 * @brief Get the named direct child
			 *
			 * This function is non-recursive and only searches the direct
			 * childs for a complete hit of the name.
			 *
			 * @param n name of the child to find.
			 */
			QTreeItem* child( QString n ) const {
				foreach( QObject* child, childobjects() ) {
					QTreeItem* item = dynamic_cast<QTreeItem*>( child );
					if ( item != 0 && n == item->objectName() )
						return item;
				}
				return 0;
			}
			/**
			 * @brief Get the named direct child of the given type
			 *
			 * This functions extends child() and returns a casted pointer. If a
			 * child with the given name and type exists. Otherwise it returns 0.
			 *
			 * @param n name of the child to find
			 */
			template <class T>
			T* child( QString n ) const {
				foreach( T* child, childobjects<T>() ) {
					if ( child->objectName() == n )
						return child;
				}
				return 0;
			}

			/**
			 * @brief Get the named item recursivly
			 *
			 * This returns the item you select by path or 0 if no item is found.
			 *
			 * @param n the path to the item you search
			 */
			QTreeItem* getNamedChild( QString n ) const {
				//qDebug() << "AbstractNamedTreeItem::getNamedChild( QString" << n << ")";
				//qDebug() << " known childs are:" << childlist();
				//qDebug() << " own name is:" << name();
				QTreeItem* item = child( n.section( "/", 0,0 ) );
				if ( item != 0 ) {
					QString path = n.section( "/", 1,-1 );
					if ( path == "" ) {
						//qDebug() << "  path is empty, returning item(" << item << ")";
						return item;
					} else {
						//qDebug() << "  returning item->getNamedChild(" << path << ")";
						return item->getNamedChild( path );
					}
				}
				return 0;
			}

			/**
			 * @brief Returns the row-number of this item inside its parent.
			 */
			int row() const {
				if ( dynamic_cast<QTreeItem*>( parent() ) )
					return dynamic_cast<QTreeItem*>( parent() )->row( this );
				return -1;
			}
			/**
			 * @brief Returns the row-number of the given child.
			 *
			 * @param child the child that is searched for.
			 */
			int row( const QTreeItem* child ) const {
				int ret = -1;
				QList<QObject*> list = childobjects();
				for ( int i=0; i<list.size(); ++i )
					if ( list[ i ] == child ) ret = i;
				return ret;
			}

			/**
			 * @brief return the parent path
			 *
			 * This checks for all the parent QTreeItem and returns their
			 * names as a unix-like-path.
			 */
			QString path() const {
				QString ret="";
				QTreeItem* item = dynamic_cast<QTreeItem*>( parent() );
				if ( item )
					ret = item->path() + "/";
				return ret + objectName();
			}

			/**
			 * @brief GUI: return a/the value to display
			 *
			 * If you implement this, always remember to call the parent
			 * guiValue() if you can't return the right thing.
			 */
			virtual QVariant guiValue() const { return QVariant(); }
			/**
			 * @brief GUI: set the value
			 *
			 * This function will be called by GUIs to change the value. You
			 * should implement it accordingly.
			 */
			virtual void guiSetValue( QVariant ) {}

			/**
			 * @brief GUI: return the supported actions
			 *
			 * These actions will get symbols/buttons to get executed.
			 *
			 * If you implement this, remember to call the parent guiActions()
			 * too:
			 *
			 * @code
			 * SomeTreeItem::guiActions() const {
			 *   return QStringList() << <your actions> << BaseClass::guiActions();
			 * }
			 * @endcode
			 */
			virtual QStringList guiActions() const { return QStringList(); }
			/**
			 * @brief GUI: execute the named action
			 *
			 * If you implemented guiActions() you will also want to implement
			 * this. If you can't handle the named action yourself, call
			 * guiExecAction() of the parent class.
			 *
			 * Of course this can also be used as a filter if you want to
			 * restrict access to actions of the parent class.
			 *
			 * The second parameter lets you interact with the user by some
			 * basic means.
			 *
			 * @param n The action to execute
			 * @param gui The gui to use for interactions
			 */
			virtual void guiExecAction( QString n, GuiServer_Interface* gui ) {}
			/**
			 * @brief GUI: return a string do be displayed instead of plain
			 * pointer-address
			 *
			 * @deprecated Usage should be replaced by guiIdentifier.
			 */
			virtual QString guiString() const __attribute__( ( deprecated ) ) {
				QString tmp = objectName();
				if ( tmp=="" )
					tmp = QString( "%1" ).arg( pint( this ) );
				return tmp;
			}
			/**
			 * @brief GUI: return different types of data used to identify the
			 * treeitem
			 *
			 * Things like the name, icon, statustip, tooltip and so on will be
			 * accessed by this.
			 *
			 * @param role is actually a combination of Qt::ItemDataRole
			 */
			virtual QVariant guiIdentifier( /*Qt::ItemDataRole*/ int role ) const {
				if ( role == Qt::DisplayRole )
					return guiString();
				return QVariant();
			}
			/**
			 * @brief GUI: the Flags for this item
			 */
			virtual Qt::ItemFlags guiFlags() const { return Qt::ItemIsEnabled; }


			/// Debugoutput
			/**
			 * This prints a tree-view of this item and its childs to qDebug.
			 * Every subclass has to implement debug() to return the internal
			 * debug-data.
			 *
			 * Call it without an argument.
			 */
			void debugPrint( QString prefix="" ) const {
				if ( prefix == "" )
					qDebug() << "TreeItem::debugPrint()";
				qDebug( "%s", QString( "%1 +%2 (childs %3)" ).arg( prefix ).arg( debug() ).arg( childs() ).toStdString().c_str() );
				foreach( QObject* obj, childobjects() ) {
					//qDebug() << " obj=" << obj;
					QTreeItem* item = 0;
					if ( ( item = dynamic_cast<QTreeItem*>( obj ) ) )
						item->debugPrint( prefix+" |" );
				}
			}

			/**
			 * @{
			 * @deprecated These will probably be removed in the near future.
			 */
			int maximumChilds() const __attribute__( ( deprecated ) ) { return -1; }
			void maximumChilds( int ) __attribute__( ( deprecated ) ) {};
			// @}

		public slots:
			/**
			 * @brief deferred deletion
			 *
			 * Use this one instead of QObject::deleteLater()!!!
			 */
			void treeDeleteLater() {
				//qDebug() << "QTreeItem::treeDeleteLater()";
				_deferredDelete = true;
				clearPointers();
				setParent( 0 );
				deleteLater();
			}

		protected:
			/**
			 * @brief Custom debug text
			 *
			 * This function has to be implemented and has to return the
			 * internal debug-data that is to be printed...
			 */
			virtual QString debug() const =0;

			/**
			 * @brief Adds new childs to this item
			 *
			 */
			bool addChild( QTreeItem* item ) {
				if ( ! childobjects().contains( item ) )
					_morechilds.append( item );

				if ( childobjects().contains( item ) ) {
					for ( std::list<PointerProtection_Base*>::iterator it = _pointers.begin(); it != _pointers.end(); ++it ) {
						QTreeItemPointer* tmp = dynamic_cast<QTreeItemPointer*>( *it );
						if ( tmp != 0 )
							tmp->newChildAdded( item );
					}
					return true;
				}

				qWarning() << "QTreeItem::addChild(" << item << ") returning false!";
				return false;
			}

		private:
			/// The additional "childs" not handled by qobjects system
			QObjectList _morechilds;
			/// Wether the object is in deferred deletion
			bool _deferredDelete;
	};

};

#endif // ABSTRACTTREEITEM_H

