/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "qtreeitemmodel.h"
#include "qtreeitemmodel.moc"

#include <QtCore/QDebug>
#include <QtCore/QSize>
#include <QtCore/QTimer>
#include <QtCore/QThread>

#include <namedparametertreeitem.h>

namespace NSCore {

	class QTreeItemModel_Root : public QTreeItem
	{
		friend class QTreeItemModel;
		private:
			QTreeItemModel_Root() {
				//maximumChilds( 1 );
			}
			QString debug() const { return "QTreeItemModel_Root"; }
	};

QTreeItemModel::QTreeItemModel( QObject* p )
	: QAbstractItemModel( p )
	, _roottreeitem( new QTreeItemModel_Root() )
	, _rootitem( new QTreeItemNode( _roottreeitem, 0 ) )
{
	qDebug() << "QTreeItemModel::QTreeItemModel(" << p << ") living in thread " << thread();
}
QTreeItemModel::~QTreeItemModel() {
	qDebug() << "QTreeItemModel::~QTreeItemModel()";
	delete _rootitem;
}

int QTreeItemModel::columnCount( const QModelIndex& parent ) const {
	//qDebug() << "QTreeItemModel::columnCount( parent=" << parent << ")";
	if ( _rootitem->isNull() )
		return 0;

	return 3;
}
int QTreeItemModel::rowCount( const QModelIndex& parent ) const {
	//qDebug() << "QTreeItemModel::rowCount( parent=" << parent << ")";
	QTreeItemNode* p = itemFromIndex( parent );
	if ( !p )
		return 0;
	//qDebug() << "QTreeItemModel::rowCount( parent=" << parent << ") will return" << p->childs();
	return p->childs();
}

QModelIndex QTreeItemModel::index( int row, int col, const QModelIndex& parent ) const {
	if ( _rootitem->isNull() )
		return QModelIndex();
	//qDebug() << "QTreeItemModel::index( row=" << row << ", col=" << col << ", parent=" << parent << ")";

	QTreeItemNode* p = itemFromIndex( parent );
	if ( row >= 0 && row < p->childs() )
		return createIndex( row, col, p->child( row ) );

	return QModelIndex();
}
QModelIndex QTreeItemModel::parent( const QModelIndex& index ) const {
	//qDebug() << "QTreeItemModel::parent( index=" << index << ")";

	QTreeItemNode* item = itemFromIndex( index );
	if ( !item )
		return QModelIndex();

	QTreeItemNode* parentItem = item->parentNode();
	if ( !parentItem )
		return QModelIndex();

	int row = parentItem->row();
	if ( row < 0 )
		return QModelIndex();
	return createIndex( row, 0, parentItem );
}

QVariant QTreeItemModel::data( const QModelIndex& index, int role ) const {
	if ( ! index.isValid() )
		return QVariant();

	QVariant ret;
	NSCore::QTreeItemNode* tmp = static_cast<NSCore::QTreeItemNode*>( index.internalPointer() );
	if ( tmp != 0 && tmp->object() != 0 ) {
		if ( index.column() == 0 ) {
			ret = tmp->object()->guiIdentifier( role );
		} else {
			if ( role == Qt::DisplayRole || role == Qt::EditRole ) {
				switch ( index.column() ) {
					case 2:
						ret = tmp->object()->guiActions();
						break;
					case 1:
						ret = tmp->object()->guiValue();
						break;
				}
			}
		}
	}
	return ret;
}

bool QTreeItemModel::setData( const QModelIndex& index, const QVariant& value, int role ) {
	NSCore::QTreeItemNode* tmp = static_cast<NSCore::QTreeItemNode*>( index.internalPointer() );
	if ( role != Qt::EditRole || index.column() != 1 || tmp->object()->guiValue().type() == QVariant::Invalid )
		return false;
	tmp->object()->guiSetValue( value );
	return true;
}

Qt::ItemFlags QTreeItemModel::flags( const QModelIndex& index ) const {
	Qt::ItemFlags ret;
	NSCore::QTreeItemNode* tmp = static_cast<NSCore::QTreeItemNode*>( index.internalPointer() );
	if ( tmp && tmp->object() )
		ret = tmp->object()->guiFlags();
	if ( index.column() != 1 )
		ret &= ~Qt::ItemIsEditable;
	return ret;
}

QVariant QTreeItemModel::headerData( int section, Qt::Orientation orientation, int role ) const {
	if ( orientation == Qt::Horizontal && role == Qt::DisplayRole ) {
		if ( section == 0 )
			return "Name";
		if ( section == 1 )
			return "Value";
		if ( section == 2 )
			return "Actions";
	}
	return QVariant();
}

bool QTreeItemModel::removeRows( int row, int count, const QModelIndex& parent ) {
	qDebug() << "QTreeItemModel::removeRows(" << row << "," << count << "," << parent << ")";
	bool ret = QAbstractItemModel::removeRows( row, count, parent );
	qDebug() << " will return" << ret;
	return ret;
}

void QTreeItemModel::setRootItem( NSCore::QTreeItem* item ) {
	qDebug() << "QTreeItemModel::setRootItem(" << item << ")";
	addToRootItem( item );
}
void QTreeItemModel::addToRootItem( NSCore::QTreeItem* item ) {
	qDebug() << "QTreeItemModel::addToRootItem(" << item << ")";
	_roottreeitem->addChild( item );
	connect( _rootitem, SIGNAL( itemDeleted( QTreeItemNode* ) ), this, SLOT( nodeDeleted( QTreeItemNode* ) ) );
	connect( _rootitem, SIGNAL( newChild( QTreeItemNode*, QTreeItemNode* ) ), this, SLOT( nodeAdded( QTreeItemNode*, QTreeItemNode* ) ) );
}

QTreeItemNode* QTreeItemModel::itemFromIndex( const QModelIndex& index ) const {
	if ( index.isValid() )
		return static_cast<QTreeItemNode*>( index.internalPointer() );
	else
		return _rootitem;
}
QModelIndex QTreeItemModel::indexFromItem( const QTreeItemNode* item ) const {
	if ( item == _rootitem )
		return createIndex( 0, 0, _rootitem );
	if ( item == 0 )
		return QModelIndex();

	QModelIndex parentindex = indexFromItem( item->parentNode() );
	return index( item->row(), 0, parentindex );
}

void QTreeItemModel::nodeDeleted( QTreeItemNode* item ) {
	qDebug() << "QTreeItemModel::nodeDeleted(" << item << ")";
	if ( thread() == QThread::currentThread() ) {
		int row = indexFromItem( item ).row();
		QModelIndex parentindex = indexFromItem( item ).parent();
		beginRemoveRows( parentindex, row, row );
		//qDebug() << " items object is" << item->object();
		//qDebug() << " Doing the real delete on" << item;
		item->treeDeleteLater();
		endRemoveRows();
	} else {
		qDebug() << "QTreeItemModel::nodeDeleted(" << item << ") will schedule the deletion for later";
		_nodesToDelete.prepend( item );
		QTimer::singleShot( 20, this, SLOT( checkForChanges() ) );
	}
}
void QTreeItemModel::nodeAdded( QTreeItemNode* newChild, QTreeItemNode* childsparent ) {
	//qDebug() << "QTreeItemModel::nodeAdded(" << newChild << "," << childsparent << ")";
	QModelIndex parentindex( indexFromItem( childsparent ) );
	int row = rowCount( parentindex );
	beginInsertRows( parentindex, row+1, row+1 );
	endInsertRows();
}

void QTreeItemModel::checkForChanges() {
	while ( ! _nodesToDelete.isEmpty() )
		nodeDeleted( _nodesToDelete.takeFirst() );
}


QTreeItemNode::QTreeItemNode( QTreeItem* item, QTreeItemNode* p )
	: QTreeItem( p )
	, QTreeItemPointer( item )
	, _deferredDeletion( false )
{
	//qDebug() << "QTreeItemNode::QTreeItemNode(" << item << "," << p << ") =" << this << " living in thread " << thread();
	if ( item )
		for ( int i=0; i<item->childs(); ++i )
			newChildAdded( item->child( i ) );

	startTimer( 10 );
}
QTreeItemNode::~QTreeItemNode() {
	//qDebug() << "QTreeItemNode::~QTreeItemNode()" << this;
}

void QTreeItemNode::itemDeleted() {
	//qDebug() << "QTreeItemNode::itemDeleted()";
	if ( thread() != QThread::currentThread() )
		_deferredDeletion = true;
	else {
		Q_ASSERT( thread() == QThread::currentThread() );
		Q_ASSERT( this != 0 );
		if ( this != 0 )
			emit itemDeleted( this );
	}
}
void QTreeItemNode::newChildAdded( QTreeItem* item ) {
	//qDebug() << "QTreeItemNode::newChildAdded(" << item << ")";
	if ( thread() != QThread::currentThread() ) {
		qWarning( "QTreeItemNode::newChildAdded() Will buffer the new event" );
		_bufferedNewItems.push_back( item );
	} else {
		if ( item == 0 )
			return;
		qDebug() << "QTreeItemNode::newChildAdded(" << item << ")";
		QTreeItemNode* tmp = new QTreeItemNode( item, this );
		connect( tmp, SIGNAL( itemDeleted( QTreeItemNode* ) ), this, SIGNAL( itemDeleted( QTreeItemNode* ) ) );
		connect( tmp, SIGNAL( newChild( QTreeItemNode*, QTreeItemNode* ) ), this, SIGNAL( newChild( QTreeItemNode*, QTreeItemNode* ) ) );
		emit newChild( tmp, this );
	}
}


QTreeItemNode* QTreeItemNode::parentNode() const {
	if ( !QTreeItem::parent() ) {
		qWarning() << "QTreeItemNode::parentNode() Got no parent anymore";
		return 0;
	}
	return dynamic_cast<QTreeItemNode*>( QTreeItem::parent() );
}

QTreeItemNode* QTreeItemNode::child( int n ) const {
	return static_cast<QTreeItemNode*>( QTreeItem::child( n ) );
}

QString QTreeItemNode::debug() const {
	if ( _p )
		return QString( "Pointer to: " );
	return "Pointer to 0!";
}

void QTreeItemNode::timerEvent( QTimerEvent* ) {
	//qDebug() << "QTreeItemNode::timerEvent( QTimerEvent* )";
	while ( !_bufferedNewItems.isEmpty() )
		newChildAdded( _bufferedNewItems.takeFirst() );

	if ( _deferredDeletion )
		itemDeleted();
}

};
