/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef QTREEITEMMODEL_H
#define QTREEITEMMODEL_H

#include <QtCore/QAbstractItemModel>
#include <QtCore/QList>
#include <pointer_template.h>
#include <qtreeitem.h>

/**
 * \defgroup module_treeitemmodel TreeItemModel
 *
 * The classes in this module combine the Model-View-Pattern used in Qt with the
 * coolness of trees of QTreeItem.
 *
 * If you have structure of childclasses of QTreeItem and want a "nice" GUI for
 * them you may dream about a way to use Qt's views. We will take several steps
 * to get that accomplished.
 *
 * \section step1 The prerequisites
 *
 * Probably you have some custom classes that inherit from NSCore::QTreeItem or
 * NSCore::QNamedTreeItem. Maybe you also incorporate
 * NSCore::NamedParameterTreeItem or NSCore::NamedQVariantTreeItem:
 * \code
 * NSCore::QTreeItem* abstracttree = <Your tree>;
 * \endcode
 *
 * \section step2 Preparing the model
 *
 * Next step is to create the model. There for we need to include
 * treeitemmodel.h and create an object of NSCore::TreeItemModel. Then we set
 * the root of the model:
 * \code
 * NSCore::TreeItemModel* model = new NSCore::TreeItemModel( <parent QObject> );
 * model->setRootPath( abstracttree );
 * \endcode
 * Note that the root of the model could be anywhere inside the tree but only
 * the item itself and its childs below are shown in the model.
 *
 * \section step3 Lets see it
 *
 * Now we can create any view and connect it to the model, but probably the
 * QTreeView makes the most sense:
 * \code
 * QTreeView* treeview = new QTreeView( <parent QWidget> );
 * treeview->setModel( model );
 * \endcode
 * At this point we will see the structure of the items but nothing more. We
 * need a way to tell the QTreeView how to display ( and edit ) our data-items.
 *
 * \section step4 Extending the GUI
 *
 * <b>To be done</b>
 *
 * @todo finish docs about the gui
 */


namespace NSCore {

	class QTreeItemNode;
	class QTreeItemModel_Root;

	//class NamedParameterTreeItem;

	/**
	 * @ingroup module_treeitemmodel
	 *
	 * @brief Provides a standard model for trees of NSCore::QTreeItem
	 *
	 * With this class it is possible to access a tree of
	 * NSCore::QTreeItem as a simple model of Qt's Model-View-Pattern.
	 * So all the different views can be used.
	 *
	 * Usage inside a qobject is like the following:
	 * @code
	 * TreeItemModel* model = new TreeItemModel( this );
	 * model->setRootPath( rootitem ); // Any QTreeItem will do...
	 * QTreeWidget* tw = new QTreeWidget( this );
	 * tw->setModel( model );
	 * @endcode
	 *
	 * @todo Support drag-and-drop to rearrange items.
	 */
	class QTreeItemModel : public QAbstractItemModel
	{
		Q_OBJECT
		public:
			/// Default constructor
			QTreeItemModel( QObject* );
			/// destructor
			~QTreeItemModel();

			/// implements QAbstractItemModel::columnCount()
			int columnCount( const QModelIndex& parent = QModelIndex() ) const;
			/// implements QAbstractItemModel::rowCount()
			int rowCount( const QModelIndex& parent = QModelIndex() ) const;

			/// implements QAbstractItemModel::index()
			QModelIndex index( int row, int col, const QModelIndex& parent = QModelIndex() ) const;
			/// implements QAbstractItemModel::parent()
			QModelIndex parent( const QModelIndex& index ) const;

			/// implements QAbstractItemModel::data()
			QVariant data( const QModelIndex& index, int role = Qt::DisplayRole ) const;
			/// implements QAbstractItemModel::setData()
			bool setData( const QModelIndex& index, const QVariant& value, int role = Qt::EditRole );
			/// implements QAbstractItemModel::headerData()
			QVariant headerData( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const;

			/// implements QAbstractItemModel::flags()
			Qt::ItemFlags flags( const QModelIndex& ) const;

			bool removeRows( int, int, const QModelIndex& = QModelIndex() );

			/**
			 * @brief Set the root-item.
			 *
			 * @deprecated use addToRootItem() instead
			 */
			void setRootItem( QTreeItem* );

			/**
			 * @brief Add an item to the root.
			 *
			 * The added item can be any QTreeItem, even one in the middle of a
			 * tree.
			 */
			void addToRootItem( QTreeItem* );

		private slots:
			/// A node was deleted
			void nodeDeleted( QTreeItemNode* );
			/// A new node was added
			void nodeAdded( QTreeItemNode* newchild, QTreeItemNode* childsparent );

			void checkForChanges();

		private:
			QTreeItemNode* itemFromIndex( const QModelIndex& ) const;
			QModelIndex indexFromItem( const QTreeItemNode* ) const;

			QTreeItemModel_Root* _roottreeitem;
			QTreeItemNode* _rootitem;
			QList<QTreeItemNode*> _nodesToDelete;
	};

	/**
	 * @ingroup module_treeitemmodel
	 * @internal
	 * @brief Internal class to sort tree-items for QTreeItemModel.
	 */
	class QTreeItemNode : public QTreeItem, public QTreeItemPointer
	{
		Q_OBJECT
		public:
			QTreeItemNode( QTreeItem*, QTreeItemNode* );
			~QTreeItemNode();

			void itemDeleted();

			void newChildAdded( QTreeItem* );

			QTreeItemNode* parentNode() const;

			QTreeItemNode* child( int n ) const;

		signals:
			/**
			 * @brief A child was deleted
			 *
			 * The QTreeItemNode* is either this item itself ( if its the one
			 * that was deleted ) or one of its childs.
			 */
			void itemDeleted( QTreeItemNode* );
			/**
			 * @brief A new child was added
			 *
			 * @param child The new item that was added
			 * @param parent The parent item the new child was added to
			 */
			void newChild( QTreeItemNode* child, QTreeItemNode* parent );

		private:
			QString debug() const;
			void timerEvent( QTimerEvent* );
			QList<QTreeItem*> _bufferedNewItems;
			bool _deferredDeletion;
	};

};


#endif // NAMEDTREEMODEL_H

