/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "diskiobuffers.h"
#include "diskiobuffers.moc"

#include <QtCore/QDebug>
#include <QtCore/QFile>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>

namespace NSDiskIO {

	AcqirisLiveReader::AcqirisLiveReader( QObject* p )
		: AbstractReader( p )
		, AbstractDiskIO<QTextStream>( QIODevice::ReadOnly )
		, _samplinginterval( 0 )
		, _channelfsr( 5 )
		, _fullscale( 256 )
		, _horizontalposition( 0 )
		, _lsb( 0 )
	{
		qDebug() << "AcqirisLiveReader::AcqirisLiveReader(" << p << ")";
	}

	void AcqirisLiveReader::readSingleDataSet() {
		if ( _stream != 0 ) {
			bool finished = false;
			int i=0;
			while ( ! _stream->atEnd() && !finished ) {
				QString line = _stream->readLine();
				if ( line[ 0 ] == '$' ) {
					QString first = line.section( " ", 0, 0 );
					QString last = line.section( " ", -1, -1 );
					if ( first == "$SAMPIVAL" )
						_samplinginterval = last.toDouble();
					if ( first == "$CHANNELFSR" )
						_channelfsr = last.toDouble();
					if ( first == "$FULLSCALE" )
						_fullscale = last.toDouble();
					if ( first == "$HORPOS" )
						_nexthorizontalposition = last.toDouble();

					if ( _horizontalposition == 0 )
						_horizontalposition = _nexthorizontalposition;

					_lsb = _channelfsr / _fullscale;

					// only stop If you have read some data:
					if ( _data.size() > 0 )
						finished = true;
				} else {
					++i;
					_data.push_back( line.toDouble() * _lsb );
				}
			}

			NSCore::AcqisitionDataSet* dataset = new NSCore::AcqisitionDataSet( 1, _data.size() );
			dataset->parameter( "horizontalposition" )->setValue( _horizontalposition );
			dataset->parameter( "samplinginterval" )->setValue( _samplinginterval );
			NSCore::ChannelDataSet* channel = dataset->channel( 0 );
			channel->parameter( "fullscale" )->setValue( _channelfsr );

			for ( int i=0; i<_data.size(); ++i ) {
				channel->setSample( i, _data[ i ] );
			}

			enqueue( dataset );

			_data.clear();
			_horizontalposition = _nexthorizontalposition;
		}
	}

	CSVWriter::CSVWriter( QObject* p ) : AbstractWriter( p ), AbstractDiskIO<QTextStream>( QIODevice::WriteOnly ) { }
	void CSVWriter::writeDataSet( NSCore::AcqisitionDataSet& data ) {
		if ( _stream != 0 ) {
			double samplinginterval = data.parameter( "samplinginterval" )->toDouble();
			double triggerdelay = data.parameter( "triggerdelay" )->toDouble();
			double horizontalposition = data.parameter( "horizontalposition" )->toDouble();
			for ( int i=0; i<data.channel( 0 )->samples(); ++i ) {
				*_stream << horizontalposition + triggerdelay + i*samplinginterval;
				for ( int j=0; j<data.channels(); ++j ) {
					*_stream << ", " << data.channel( j )->sample( i );
				}
				*_stream << endl;
			}
			*_stream << endl << endl;
		}
	}

	EPOSWriter::EPOSWriter( QObject* p ) : AbstractWriter( p ), AbstractDiskIO<QTextStream>( QIODevice::WriteOnly ) { }
	void EPOSWriter::writeDataSet( NSCore::AcqisitionDataSet& data ) {
		if ( _stream != 0 ) {
			*_stream << data.parameter( "samplinginterval" )->toDouble()
			         << " "
			         << data.parameter( "triggerdelay" )->toDouble()
			         << " "
			         << data.parameter( "horizontalposition" )->toDouble()
			         << " "
			         << data.parameter( "timestamp" )->toDouble()
			         << " "
			         << data.channels()
			         << " "
			         << data.channel( 0 )->samples()
			         << " ";
			for ( int i=0; i<data.channels(); ++i ) {
				*_stream << data.channel( i )->parameter( "fullscale" )->toDouble()
				         << " "
				         << data.channel( i )->parameter( "offset" )->toDouble()
				         << " "
				         ;
				for ( int j=0; j<data.channel( i )->samples(); ++j ) {
					*_stream << data.channel( i )->sample( j ) << " ";
				}
			}
			*_stream << endl;
		}
	}

	EPOSReader::EPOSReader( QObject* p ) : AbstractReader( p ), AbstractDiskIO<QTextStream>( QIODevice::ReadOnly ) { }
	void EPOSReader::readSingleDataSet() {
		if ( _stream != 0 && ! _stream->atEnd() ) {

			double samplinginterval, triggerdelay, horizontalposition, timestamp;
			int channels, samples;
			*_stream >> samplinginterval; _stream->skipWhiteSpace();
			*_stream >> triggerdelay; _stream->skipWhiteSpace();
			*_stream >> horizontalposition; _stream->skipWhiteSpace();
			*_stream >> timestamp; _stream->skipWhiteSpace();
			*_stream >> channels; _stream->skipWhiteSpace();
			*_stream >> samples; _stream->skipWhiteSpace();

			NSCore::AcqisitionDataSet* data = new NSCore::AcqisitionDataSet( channels, samples );
			data->parameter( "samplinginterval" )->setValue( samplinginterval );
			data->parameter( "triggerdelay" )->setValue( triggerdelay );
			data->parameter( "horizontalposition" )->setValue( horizontalposition );
			data->parameter( "timestamp" )->setValue( timestamp );

			double fullscale, offset, sample;
			for ( int i=0; i<channels; ++i ) {
				NSCore::ChannelDataSet* channel = data->channel( i );

				*_stream >> fullscale; _stream->skipWhiteSpace();
				*_stream >> offset; _stream->skipWhiteSpace();
				channel->parameter( "fullscale" )->setValue( fullscale );
				channel->parameter( "offset" )->setValue( offset );

				for ( int j=0; j<samples; ++j ) {
					*_stream >> sample; _stream->skipWhiteSpace();
					channel->setSample( j, sample );
				}

			}

			enqueue( data );
		}
	}
	void EPOSReader::setFile( QString file ) {
		AbstractDiskIO<QTextStream>::setFile( file );
	}
	QString EPOSReader::filename() const {
		return AbstractDiskIO<QTextStream>::filename();
	}

	CombinedCSV::CombinedCSV( QObject* p ) : AbstractWriter( p ), AbstractDiskIO<QTextStream>( QIODevice::WriteOnly ) {
	}

	void CombinedCSV::writeDataSet( NSCore::AcqisitionDataSet& data ) {
		if ( _stream != 0 ) {
			double samplinginterval = data.parameter( "samplinginterval" )->toDouble();
			double horizontalposition = data.parameter( "horizontalposition" )->toDouble() + data.parameter( "triggerdelay" )->toDouble();
			for ( int i=0; i<data.channel( 0 )->samples(); ++i ) {
				*_stream << horizontalposition + i*samplinginterval;
				for ( int j=0; j<data.channels(); ++j ) {
					*_stream << ", " << data.channel( j )->sample( i );
				}
				*_stream << endl;
			}
			*_stream << endl << endl;
		}
	}
}

