/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef EVALUATION_PATH_H
#define EVALUATION_PATH_H

#include <QtCore/QObject>
#include <QtCore/QDebug>
#include <QtCore/QList>
#include <QtCore/QVariant>
#include <QtXml/QDomNode>

#include <datasetchain.h>
#include <namedparametertreeitem.h>
#include <pointer_template.h>
#include <thread_template.h>

class Analyzer;
class Modifier;
class Filter;

class QTimer;

namespace NSEvaluation {

	class PluginItem;

	/**
	 * @brief The path of evaluation
	 *
	 * While this has to be connected to an AcqisitionDataSetBuffer to get its
	 * input-data, it is an AcqisitionDataSetBuffer itself which sends out the
	 * evaluated data.
	 *
	 * To have an own thread for the evaluationpath use NSCore::Thread_Template
	 */
	class EvaluationPath : public NSCore::DataSetChainMember
	{
		Q_OBJECT

		public:
			/// constructor
			EvaluationPath( QObject* =0, NSCore::QTreeItem* =0 );
			/// destructor
			~EvaluationPath();

			// TreeItem
			QString type() const { return "EvaluationPath"; }
			QString name() const { return ""; }

			/// Access the configuration
			NSCore::NamedParameterTreeItemPointer config() {
				return NSCore::NamedParameterTreeItemPointer( _config );
			}

			Q_SCRIPTABLE QString toXML() const;

		public slots:
			/**
			 * @brief Create the PluginItems from a XML-String
			 */
			void parseXML( const QString& );

			/**
			 * @brief Create the PluginItems from a QDomeNode
			 *
			 * The DomNode has to be a &lt;evaluationpath /&gt;.
			 */
			void parseXML( QDomNode );

			/**
			 * Create new plugins
			 */
			PluginItem* newPlugin( QString plugintype, QTreeItem* parent =0 );

			/**
			 * Delete existing plugins
			 */
			bool deletePlugin( PluginItem* );

			QStringList guiActions() const {
				return QStringList( "Add" ) << QTreeItem::guiActions();
			}
			QVariant guiIdentifier( int ) const;

			void guiExecAction( QString action, NSCore::GuiServer_Interface* gui );
			void execAdd( NSCore::QTreeItem*, NSCore::GuiServer_Interface* );
			void execRemove( NSCore::QTreeItem*, NSCore::GuiServer_Interface* );


		signals:
			/**
			 * @brief some plugins have changed.
			 *
			 * This signal is emitted when the plugins have change. It is used
			 * for the gui to update the representation of the plugin-tree.
			 */
			void pluginsChanged();

			/**
			 * @brief inform the gui ( or a frontend ) of a new plugin
			 */
			void newPlugin( NSEvaluation::PluginItem* plugin, NSCore::QTreeItem* parent );

		private slots:
			/** @internal */
			void checkBuffers();

		private:
			/** @internal */
			bool dataArrived( NSCore::AcqisitionDataSet* );

			/** @internal */
			void internalParseXML( QDomNode node, NSCore::QTreeItem* parent );

			QString debug() const { return "EvaluationPath"; }

			NSCore::NamedParameterTreeItem* _config;

			QMap<QString,PluginItem*> _pluginsToAdd;
			QList<PluginItem*> _pluginsToDelete;
			QList<QDomNode> _xmlToParse;

			QTimer* _timer;
	};

	typedef NSCore::Thread_Template<EvaluationPath> EvaluationThread;

};

#endif // EVALUATION_PATH_H

