/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "osciwidget_v3.h"
#include "osciwidget_v3.moc"

#include "osciwidget_v3_private.h"
#include "osciwidget_v3_private.moc"

#include <math.h>
#include <QtCore/QMutex>
#include <QtCore/QMutexLocker>
#include <QtGui/QPainter>
#include <QtGui/QVBoxLayout>
#include <QtGui/QInputDialog>
#include <QtGui/QAction>
#include <QtGui/QMenu>
#include <QtGui/QContextMenuEvent>
#include <QtCore/QTimer>

OsciWidgetV3_ChainMember::OsciWidgetV3_ChainMember( QObject* p ) : NSCore::DataSetChainMember( p ) {
	_slowdown.start();
}
bool OsciWidgetV3_ChainMember::dataArrived( NSCore::AcqisitionDataSet* d ) {
	//qDebug() << "OsciWidgetV3_ChainMember::dataArrived(" << d << ")";
	if ( _slowdown.elapsed() > 200 ) {
		emit data( *( d ) );
		_slowdown.restart();
	}
	return true;
}

OsciWidgetV3::OsciWidgetV3( QWidget* p )
	: QWidget( p )
	, _layout( new QVBoxLayout( this ) )
	, _options( new OsciWidgetV3_Options( this ) )
	, _chainmember( new OsciWidgetV3_ChainMember( this ) )
{
	connect( _chainmember, SIGNAL( data( NSCore::AcqisitionDataSet& ) ), this, SLOT( dataAvailable( NSCore::AcqisitionDataSet& ) ) );

	setMinimumSize( 100,100 );
	_layout->addWidget( _options, 1 );
	connect( _options->spinboxChannels, SIGNAL( valueChanged( int ) ), this, SLOT( setChannels( int ) ) );
	setChannels( _options->spinboxChannels->value() );

}
OsciWidgetV3::~OsciWidgetV3() {
}

void OsciWidgetV3::dataAvailable( NSCore::AcqisitionDataSet& data ) {
	//qDebug() << "OsciWidgetV3::dataAvailable( data )";
	for ( int i=0; i<_oscis.size(); ++i ) {
		_oscis[ i ]->data( &data );
	}
}

void OsciWidgetV3::setChannels( int n ) {
	while ( _oscis.size() > n ) {
		delete _oscis.takeLast();
	}
	while ( _oscis.size() < n ) {
		_oscis.push_back( new OsciWidgetV3_Osci( this ) );
		_layout->addWidget( _oscis.last(), 255 );
		_oscis.last()->setChannel( _oscis.size()-1 );
	}
}

NSCore::DataSetChainMember* OsciWidgetV3::chainMember() { return _chainmember; }



OsciWidgetV3_Osci::OsciWidgetV3_Osci( QWidget* p )
	: QWidget( p )
	, fullscale( 0 )
	, offset( 0 )
	, _channel( 0 )
	, _generation( 0 )
	, _mutex( new QMutex() )
{
	QAction* channelselect = new QAction( "Select Channel...", this );
	connect( channelselect, SIGNAL( triggered() ), this, SLOT( enterChannel() ) );
	addAction( channelselect );
	QAction* generationselect = new QAction( "Select Generation...", this );
	connect( generationselect, SIGNAL( triggered() ), this, SLOT( enterGeneration() ) );
	addAction( generationselect );
}
OsciWidgetV3_Osci::~OsciWidgetV3_Osci() {
	delete _mutex;
}

void OsciWidgetV3_Osci::data( NSCore::AcqisitionDataSet* data ) {
	if ( ! data->channels() > _channel )
		_channel = data->channels() - 1;

	NSCore::ChannelDataSet* channel = data->channel( _channel );
	if ( channel ) {

		QMutexLocker locker( _mutex );

		int generation = _generation;
		if ( channel->generations() <= generation )
			generation = channel->currentGeneration();

		fullscale = channel->parameter( "fullscale" )->toDouble();
		offset = channel->parameter( "offset" )->toDouble();
		if ( fullscale == 0 ) {
			double a = channel->result( "max", generation )->toDouble();
			double b = channel->result( "min", generation )->toDouble();
			fullscale = a - b;
			offset = ( a+b ) / 2;
		}

		_data.clear();
		int inc = 1;
		if ( channel->samples() > width() )
			inc = channel->samples() / width();

		for ( int i=0; i<channel->samples(); i+=inc )
			_data.push_back( channel->sample( i, generation ) );

		_addons_x[ "Minimum at" ] = 0;
		_addons_x[ "Fraction Point At" ] = 0;
		_addons_x[ "Threshold At" ] = 0;
		_addons_y[ "Minimum" ] = 0;
		_addons_y[ "zeroline" ] = 0;
		_addons_y[ "Fraction Point" ] = 0;
		if ( channel->result( "interpolated threshold at", generation ) != 0 )
			_addons_x[ "Threshold at" ] = channel->result( "interpolated threshold at", generation )->toDouble();
		if ( channel->result( "interpolated minimum at", generation ) != 0 )
			_addons_x[ "Minimum at" ] = channel->result( "interpolated minimum at", generation )->toDouble();
		if ( channel->result( "interpolated fraction point at", generation ) != 0 )
			_addons_x[ "Fraction Point At" ] = channel->result( "interpolated fraction point at", generation )->toDouble();
		if ( channel->result( "interpolated minimum", generation ) != 0 )
			_addons_y[ "Minimum" ] = channel->result( "interpolated minimum", generation )->toDouble();
		if ( channel->result( "zeroline", generation ) != 0 )
			_addons_y[ "zeroline" ] = channel->result( "zeroline", generation )->toDouble();
		if ( channel->result( "interpolated fraction point", generation ) != 0 )
			_addons_y[ "Fraction Point" ] = channel->result( "interpolated fraction point", generation )->toDouble();
	}
	update();
}

void OsciWidgetV3_Osci::paintEvent( QPaintEvent* ) {
	QMutexLocker locker( _mutex );
	QPainter p( this );

	p.save();

	double dx = width() / double( _data.size() );
	double dy = - height() / fullscale;

	p.translate( 0, height()/2 + dy*offset );

	/*p.setPen( QColor( 150,150,150 ) );
	p.drawText( QPointF( 0, 10 ),
			QString( "Fullscale: %1, Offset: %2, results in [%3,%4], dy: %5" )
			.arg( fullscale )
			.arg( offset )
			.arg( ( -offset - fullscale/2 ) * dy )
			.arg( ( -offset + fullscale/2 ) * dy )
			.arg( dy )
			);*/

	p.setPen( QColor( 0,0,0 ) );

	p.drawLine( 0, 0, width(), 0 );

	if ( _data.size() > 0 ) {
		QPointF oldpoint( 0, _data[ 0 ] * dy );
		QPointF newpoint;
		for ( int i=1; i<_data.size(); ++i ) {
			newpoint = QPointF(  i*dx, _data[ i ] * dy );
			p.drawLine( oldpoint, newpoint );
			oldpoint = newpoint;
		}
	}

	p.setPen( QColor( 100,100,255 ) );
	{
		QMapIterator<QString,double> it( _addons_x );
		int i=0;
		while ( it.hasNext() ) {
			it.next();
			if ( it.value() != 0 ) {
				int x = int( dx * it.value() );
				p.drawLine( x, int( -fullscale*dy ), x, int( fullscale*dy ) );
				p.drawText( x+10, 30 - i*20, QString( "%1: %2" ).arg( it.key() ).arg( it.value() ) );
				++i;
			}
		}
	}
	{
		QMapIterator<QString,double> it( _addons_y );
		while ( it.hasNext() ) {
			it.next();
			if ( it.value() != 0 ) {
				int y = int( dy * it.value() );
				p.drawLine( 0, y, width(), y );
				p.drawText( 10, y-10, QString( "%1: %2" ).arg( it.key() ).arg( it.value() ) );
			}
		}
	}

	p.restore();
	p.drawRect( rect().adjusted( 0,0, -1, -1 ) );
}

void OsciWidgetV3_Osci::setChannel( int n ) { _channel = n; }

void OsciWidgetV3_Osci::contextMenuEvent( QContextMenuEvent* ev ) {
	QMenu::exec( actions(), ev->globalPos() );
}

void OsciWidgetV3_Osci::enterChannel() {
	_channel = QInputDialog::getInteger( this,
			"Select the new channel to show",
			"Channel to show:",
			_channel, 0 );
}
void OsciWidgetV3_Osci::enterGeneration() {
	_generation = QInputDialog::getInteger( this,
			"Select the new generation to show",
			"Generation to show:",
			_generation, 0 );
}


OsciWidgetV3_Options::OsciWidgetV3_Options( QWidget* p )
	: QFrame( p )
{
	setupUi( this );
}

