/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ACQIRIS_H
#define ACQIRIS_H

#include <QtCore/QObject>
#include <QtCore/QStringList>
#include <QtCore/QVariant>
#include <QtCore/QTime>
#include <datasetchain.h>
#include <namedparametertreeitem.h>

#ifdef HAVE_ACQIRIS
#include "AcqirisDataTypes.h"
#else
typedef qint32 ViInt32;
typedef quint32 ViUInt32;
typedef ViUInt32 ViSession;
typedef double ViReal64;
#endif

/**
 * @brief Namespace for the Acqiris-related things
 *
 * This namespace contains the acqiris-related classes, interfaces and functions.
 *
 * @todo Simplify the Digitizer and its channels
 */

namespace NSAcqiris {

	class Digitizer;

	/// Init all the crades
	/**
	 * Initializes all the crades and gives access to all the found digitizers
	 */
	class Acqiris : public NSCore::QTreeItem {
		Q_OBJECT
		public:
			/// Inits Acqiris
			Acqiris( QObject* =0 );
			/// Stops all Acqiris-things.
			// Should not be deleted as long as you want to access digitizers...
			// Deletes all the digitizers on exit.
			~Acqiris();

			/// Returns the number of found digitizers
			int digitizers() { return _digitizers.size(); }

			/// Returns a pointer to the specified digitizer
			Digitizer* digitizer( int );

		private:
			QString debug() const { return "acqiris-hw-access"; }
			std::list<Digitizer*> _digitizers;
	};

	/**
	 * @brief One digitizer
	 * Represents a digitizer in all its beauty
	 */
	class Digitizer : public NSCore::DataSetProducer {
		Q_OBJECT
		friend class Acqiris;

		public:
			/// wether to do single or continuous acquisations
			bool singleShot();

			/// how many channels are there?
			int channels();

			/// how many external trigger are there?
			int externaltriggers();

			/**
			 * @brief access the configuration-tree
			 */
			NSCore::NamedParameterTreeItemPointer config() { return _c; }

			QStringList triggersources();
			static QStringList triggerslopes() {
				return QStringList()<<"Positive"<<"Negative"<<"Out of Window"<<"Into Window"<<"HF Divide";
			}
			static QStringList triggercouplings() {
				return QStringList()<<"DC"<<"AC"<<"HF Reject"<<"DC 50Ohm"<<"AC 50Ohm";
			}

			/// Returns a list of possible FullScales
			static QStringList fullscales() {
				return QStringList()<<"50mV"<<"100mV"<<"200mV"<<"500mV"<<"1V"<<"2V"<<"5V";
			}

			/// Returns a list of possible couplings
			static QStringList couplings() {
				return QStringList()/*<<"Ground"*/<<"DC 1MOhm"<<"AC 1MOhm"<<"DC 50Ohm"<<"AC 50Ohm";
			}

			/// Returns a list of possible bandwidths
			static QStringList bandwidths() {
				return QStringList()<<"no limit"<<"25MHz"<<"700MHz"<<"200MHz"<<"20MHz"<<"35MHz";
			}

			static QStringList LEDColors() {
				return QStringList()
					<< "Off"
					<< "Green"
					<< "Red"
					<< "Yellow"
					;
			}

			QString toXML() const { return "<digitizer />"; }

		public slots:
			/// do single or continuous acquirations
			void singleShot( bool );

			/// Starts the acquisition
			void startAcqisition();
			/// Stops the acquisition (either by forcedTrigger or not...)
			void stopAcqisition();

			/// Stops a current Acquisition and does a full calibration
			void calibrate();

			/// Set the parameters
			void setSamplingInterval( const QString& );

			/// Parameters for the trigger
			void triggersource( const QString& );
			void triggerslope( const QString& );
			void triggercoupling( const QString& );

			void setLEDColor( const QString& );

		signals:
			/// an acquisition was started
			void acquisitionStarted();
			/// the acquisition stopped regularly
			void acquisitionStopped();
			/// the acquisition stopped because of a timeout
			void acquisitionStoppedTimeout();

		private slots:
			/// [Internal] check wether the the acquisition is finished and read the data
			/**
			 * @internal
			 */
			void checkForData();

			void scheduleHorizontalConfig();
			void scheduleTriggerConfig();
			void scheduleVerticalConfig();

			void triggerLevelIntern( double );
			void triggerLevelExtern( double );
		private:
			/// Inits the digitizer
			/*
			 * Checks for a digitizer and initializes all the channels
			 */
			Digitizer( ViSession, Acqiris*, QString );
			~Digitizer();

			/// Horizontal configuration
			void configHorizontal();
			/// Trigger configuration
			void configTrigger();
			/// Vertical configuration
			void configVertical();

			/// Converts a triggerpattern to a long
			long triggerSourceToPattern( int ) const;

			/// Converts a fullscale-string to double
			double fullscale( QString ) const;

			QString debug() const;

			/// Variables
			ViSession _instrumentID;
			NSCore::NamedParameterTreeItemPointer _c;
			bool _singleShot, _stopped, _horizontalConfigNeeded, _triggerConfigNeeded, _verticalConfigNeeded;
			QList<ViReal64*> _waveFormArrays;
			int _waveFormArraySize;

			QTime _time_since_last_start;
	};

};

#endif // ACQIRIS_H

