/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "scriptenvironment.h"
#include "scriptenvironment.moc"

#include "scriptsnippet.h"

// Ugly hack to have QThread::*sleep()...
#define protected public
#include <QtCore/QThread>
#undef protected

#include <QtCore/QDebug>
#include <QtCore/QSettings>
#include <QtCore/QCoreApplication>
#include <QtScript/QScriptEngine>

using namespace NSScript;

QScriptValue sleep( QScriptContext* /*context*/, QScriptEngine* engine )
{
	QThread::currentThread()->msleep( 250 );
	QCoreApplication::processEvents( 0, 250 );
	return engine->nullValue();
}

ScriptEnvironment::ScriptEnvironment( QObject* p )
	: QObject( p )
	, _engine( new QScriptEngine( this ) )
	, _settings( new QSettings( "eposlab", "scriptenvironment", this ) )
{
	_settings->beginGroup( "script-restore" );
	int snippets = _settings->value( "scriptsnippets", int( 0 ) ).toInt();
	for ( int i=0; i<snippets; ++i ) {
		qDebug() << " Restoring snippet" << i;
		_snippetsToRestore_t snip;
		snip.name = _settings->value( QString( "scriptsnippet%1_name" ).arg( i ), QString() ).toString();
		snip.data = _settings->value( QString( "scriptsnippet%1" ).arg( i ), QString() ).toString();
		snip.savelocation = ScriptSnippet::SystemWide;
		_snippetsToRestore.append( snip );
		emit snippetsToRestoreAvailable();
	}

	_engine->globalObject().setProperty( "sleep", _engine->newFunction( sleep ) );
}
ScriptEnvironment::~ScriptEnvironment() {
	qDebug() << "ScriptEnvironment::~ScriptEnvironment()";
	systemSave();
	while ( !_snippets.isEmpty() )
		delete _snippets.takeFirst();
	
	qDebug() << "ScriptEnvironment::~ScriptEnvironment() finished";
}

void ScriptEnvironment::addQObject( QObject* obj ) {
	static int c =0;
	if ( obj->objectName() == "" )
		obj->setObjectName( QString( "scriptobject%1" ).arg( c++ ) );

	_engine->globalObject().setProperty( obj->objectName(), _engine->newQObject( obj ) );
	_objects.append( obj );
}

bool ScriptEnvironment::hasUnrestoredSnippets() const { return !_snippetsToRestore.isEmpty(); }

ScriptSnippet* ScriptEnvironment::restoreFirstSnippet( QWidget* p ) {
	//qDebug() << "ScriptEnvironment::restoreFirstSnippet(" << p << ")";
	_snippetsToRestore_t snip = _snippetsToRestore.takeFirst();
	ScriptSnippet* tmp = newSnippet( p );
	tmp->setCode( snip.data );
	tmp->setName( snip.name );
	tmp->setSaveLocation( ScriptSnippet::SaveLocation( snip.savelocation ) );
	return tmp;
}

ScriptSnippet* ScriptEnvironment::newSnippet( QWidget* p ) {
	ScriptSnippet* tmp = new ScriptSnippet( p );
	connect( tmp, SIGNAL( execScript( QString ) ), this, SLOT( execCode( QString ) ) );
	connect( tmp, SIGNAL( destroyed( ScriptSnippet* ) ), this, SLOT( snippetDestroyed( ScriptSnippet* ) ) );
	connect( tmp, SIGNAL( systemSaveRequest() ), this, SLOT( systemSave() ) );
	_snippets.append( tmp );
	tmp->show();
	return tmp;
}

void ScriptEnvironment::execCode( QString code ) {
	qDebug() << "ScriptEnvironment::execCode(" << code << ")";
	_engine->evaluate( code );
	while ( _engine->hasUncaughtException() )
		qDebug() << _engine->uncaughtExceptionBacktrace().join( "\n" ) << endl << _engine->uncaughtException().toString();
}

void ScriptEnvironment::snippetDestroyed( ScriptSnippet* obj ) {
	qDebug() << "ScriptEnvironment::snippetDestroyed(" << obj << ")";
	_snippets.removeAll( obj );
}

void ScriptEnvironment::systemSave() {
	qDebug() << "ScriptEnvironment::systemSave()";
	int i = 0;
	foreach( ScriptSnippet* snip, _snippets )
		if ( snip->getSaveLocation() == ScriptSnippet::SystemWide ) {
			_settings->setValue( QString( "scriptsnippet%1" ).arg( i ), snip->saveCode() );
			_settings->setValue( QString( "scriptsnippet%1_name" ).arg( i ), snip->name() );
			i++;
		}
	_settings->setValue( "scriptsnippets", i );
}

QString ScriptEnvironment::toXML() const {
	QString ret;
	foreach( ScriptSnippet* snip, _snippets )
		if ( snip->getSaveLocation() == ScriptSnippet::PathXML )
			ret += QString( "\t<snippet name=\"%1\">%2</snippet>" ).arg( snip->name() ).arg( snip->saveCode() );
	if ( ret.isEmpty() )
		ret = "<scripts />\n";
	else
		ret = "<scripts>\n" + ret + "</scripts>\n";
	qDebug() << "ScriptEnvironment::toXML() returning:" << ret;
	return ret;
}

void ScriptEnvironment::fromXML( QString xml ) {
	QDomDocument doc( "scripts-document" );
	doc.setContent( xml );
	fromXML( doc.documentElement() );
}
void ScriptEnvironment::fromXML( QDomNode node ) {
	qDebug() << "ScriptEnvironment::fromXML(" << &node << ")";
	if ( node.nodeName() == "scripts" ) {
		qDebug() << " node is a scripts-element";
		for ( QDomNode item = node.firstChild(); !item.isNull(); item = item.nextSibling() ) {
			qDebug() << " found sibling";
			fromXML( item );
		}
	}
	if ( node.nodeName() == "snippet" ) {
		qDebug() << " node is a snippet";
		QString name;
		if ( node.attributes().contains( "name" ) )
			name = node.attributes().namedItem( "name" ).toAttr().value();
		QString prog = node.firstChild().toText().data();
		if ( !name.isEmpty() && !prog.isEmpty() ) {
			//_snippetsToRestore[ name ] = prog;
			_snippetsToRestore_t snip;
			snip.name = name;
			snip.data = prog;
			snip.savelocation = ScriptSnippet::PathXML;
			_snippetsToRestore.append( snip );
			emit snippetsToRestoreAvailable();
		}
	}
}

