/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RESULTTOSPECTRUM_H
#define RESULTTOSPECTRUM_H

#include <datasetchain.h>
#include <abstracttreeitem.h>
#include <mathexpression.h>
#include <QtCore/QObject>
#include <QtXml/QDomNode>

class ResultsToSpectrum;
class AbstractSpectrumModel;

class QTimer;

/**
 * @brief A path of expressions together with a spectrummodel
 *
 * @todo docs
 */
class ResultAction : public NSCore::MathExpression
{
	Q_OBJECT
	friend class ResultsToSpectrum;
	protected:
		ResultAction( QString name, ResultsToSpectrum* parent );
		~ResultAction();

		void doYourJobOnData( NSCore::AcqisitionDataSet& );

		/// Dummy-implementation to fulfill NSCore::MathExpression
		double value( NSCore::AcqisitionDataSet& ) { return 0; }

	public:

		QString guiString() const { return QString( "ResultAction: %1" ).arg( objectName() ); }

		QStringList guiActions() const { return QTreeItem::guiActions() << "Remove" << "Add Exp."; }
		void guiExecAction( QString, NSCore::GuiServer_Interface* );

		void deleteMe( NSCore::MathExpression*, NSCore::GuiServer_Interface* );
		void addChild( NSCore::MathExpression*, NSCore::GuiServer_Interface* );

		QString toXML() const;

	private:
		QString debug() const { return QString( "ResultAction: %1" ).arg( objectName() ); }
		AbstractSpectrumModel* _model;
};

/**
 * @brief Sends results into a spectrummodel
 *
 * This class sends results into a SpectrumModel for further analysis and
 * storage.
 *
 * It will also be expanded to store into other models as well ( network-send,
 * sql-database, etc ).
 *
 * The results to be written into the storage can be defined by an own language
 * containing result,add,neg,smb,precision. Use and explore it online by
 * accessing this objects via TreeItemModel and QTreeView...
 */
class ResultsToSpectrum : /*public NSCore::QTreeItem,*/ public NSCore::DataSetChainMember
{
	Q_OBJECT
	public:
		/// constructor
		ResultsToSpectrum( QObject* =0 );
		/// destructor
		~ResultsToSpectrum();

	public slots:
		/**
		 * @brief Removes all paths
		 */
		void removeAllPaths();

		/**
		 * @brief Removes a specific ResultAction
		 *
		 * This is called by ResultAction in response to a user interaction.
		 *
		 * The item isn't removed directly (unless in a single-threaded app) but
		 * scheduled for removal upon next timerEvent().
		 */
		void removePath( ResultAction*, NSCore::GuiServer_Interface* );
		/**
		 * @brief Removes a specific NSCore::MathExpression
		 *
		 * This is called by ResultAction in response to a user interaction.
		 *
		 * The item isn't removed directly (unless in a single-threaded app) but
		 * scheduled for removal upon next timerEvent().
		 */
		void removePath( NSCore::MathExpression*, NSCore::GuiServer_Interface* );

		/**
		 * @brief Access the created SpectrumModels by name
		 */
		AbstractSpectrumModel* spectrum( QString name );
		/**
		 * @brief returns a list of the current spectra
		 */
		QStringList spectra() const;

		//@{
		/**
		 * @brief Create path and spectrum by XML
		 */
		void addPath( QString xml );
		void addPath( QDomNode node );
		//@}

		//@{
		/**
		 * @brief Add new Expressions
		 *
		 * This is called by ResultAction in response to a user interaction.
		 *
		 * The new item isn't created directly but scheduled for creation upon
		 * next timerEvent().
		 */
		void addExpression( ResultAction* paction, NSCore::GuiServer_Interface* );
		void addExpression( NSCore::MathExpression* pexp, NSCore::GuiServer_Interface* );
		//@}

		/**
		 * @brief Save to XML
		 *
		 * This functions saves everything needed for restore into an xml-file
		 */
		QString toXML() const;

		//@{
		/**
		 * @brief name and actions for the gui
		 */
		QStringList guiActions() const { return NSCore::QTreeItem::guiActions() << "Add Spectrum"; }
		void guiExecAction( QString, NSCore::GuiServer_Interface* );
		QVariant guiIdentifier( int ) const;
		//@}

	signals:
		/**
		 * @brief The number of spectra changed
		 *
		 * This is emitted when a new spectrum was created...
		 */
		void numberOfSpectraChanged();
		/**
		 * @brief A spectrum is scheduled for deletion
		 *
		 * Upon this signal all the guis connecting to this spectrum should be
		 * deleted. Otherwise a segfault might happen.
		 *
		 * The Spectrum isn't deleted until next timerEvent().
		 */
		void spectrumDeleted( QString );

	private slots:
		/**
		 * @internal
		 *
		 * @brief execute scheduled actions
		 *
		 * To provide thread-safety some actions have to be scheduled into a
		 * queue to be save. These queues are checked and emptied here.
		 */
		void checkBuffers();

	private:

		bool dataArrived( NSCore::AcqisitionDataSet* );

		/**
		 * @internal
		 *
		 * @brief Add a new expression
		 */
		void internalNewExpression( NSCore::QTreeItem* p, NSCore::GuiServer_Interface* );

		/// debug for NSCore::QTreeItem
		QString debug() const { return "ResultsToSpectrum"; }

		/**
		 * @internal
		 *
		 * @brief parse the XML-tree and create new items
		 *
		 * This function is thread-safe by buffering in _bufferedpaths.
		 *
		 * @see checkBuffers()
		 */
		NSCore::MathExpression* internalParseXML( QDomNode, NSCore::QTreeItem* parent=0 );

		//@{
		/**
		 * @internal
		 *
		 * @brief Buffered actions
		 *
		 * These are actions scheduled for reaction in timerEvent()
		 */
		QList<QDomNode> _bufferedpaths;
		QList<ResultAction*> _actionsToRemove;
		QList<NSCore::MathExpression*> _expressionsToRemove;
		QMap<NSCore::QTreeItem*,QString> _ExpressionsToAdd;
		QList< QPair<QDomNode,NSCore::QTreeItem*> > _internalParseXMLBuffer;
		//@}

		QTimer* _timer;

};

#endif // RESULTTOSPECTRUM_H

